from enum import Enum, auto

from pih.tools import j
from pih.consts import BARCODE
from pih.consts.file import FILE
from pih.collections import PolibaseDocumentDescription


class POLIBASE:
    
    DEVELOPER_AD_USER_LOGIN: str = "zhdanov.o"

    NAME: str = "Polibase"
    SERVICE_NAME: str = "OracleServiceORCL"
    PROCESS_NAME: str = "Polibase ODAC"

    DBMS_OUTPUT: str = "dbms_output"

    DOCTEMPLETS_LAST_OPERATED_ID: str = "doctemplete_last_operated_id"
    
    class VERSION:
        
        ORACLE: str = "11.2.0.1.0"
        VALUE: str = "23.1104.04.02"

    class NAMES:

        PERSON_PIN: str = "polibase_person_pin"
        PERSON_CARD_REGISTRY_FOLDER: str = "polibase_person_card_registry_folder"

    PRERECORDING_PIN: int = 10
    PERSON_MINIMAL_PIN: int = 100
    RESERVED_TIME_A_PIN: int = 5
    RESERVED_TIME_B_PIN: int = 6
    RESERVED_TIME_C_PIN: int = 7
    EMPTY_VALUE: str = "xxxxx"
    EMPTY_EMAIL_VARIANTS: list[str] = ["нет", "-", "no"]
    ASK_EMAIL_LOCALY_VARIANTS: list[str] = ["?"]
    TELEPHONE_NUMBER_COUNT: int = 4

    CARD_REGISTRY_FOLDER_QR_CODE_COUNT: int = 2

    CARD_REGISTRY_FOLDER_NAME_CHECK_PATTERN: list[str] = ["п", "т"]

    """
    145 - Средний Медицинский Персонал
    300 - Реанимация
    361 - Операционная блок
    421 - СМП
    221 -
    229 -
    """

    CABINET_NUMBER_EXCLUDED_FROM_VISIT_RESULT: list[int] = [
        145,
        221,
        229,
        300,
        361,
        421,
    ]

    # 147 - УЗИ
    # 201 - ЭНДОСКОПИЯ
    # 202 - МРТ
    # 203 - КТ
    # 204 - Рентген (X-ray)

    class AppointmentServiceGroupId(Enum):
        ULTRASOUND = 147
        ENDOSCOPY = 201
        MRI = 202
        CT = 203
        X_RAY = 204

    APPOINTMENT_SERVICE_GROUP_NAME: dict[Enum, str] = {
        AppointmentServiceGroupId.ULTRASOUND: "ультразвуковое исследование",
        AppointmentServiceGroupId.ENDOSCOPY: "эндоспопичекое исследование",
        AppointmentServiceGroupId.MRI: "МРТ исследование",
        AppointmentServiceGroupId.CT: "КТ исследование",
        AppointmentServiceGroupId.X_RAY: "рентген исследование",
    }

    STATUS_EXCLUDE_FROM_VISIT_RESULT: list[int] = [63]

    TELEGRAM_BOT_URL: str = "https://t.me/pacifichospital_bot"

    REVIEW_ACTION_URL: str = "https://forms.gle/qriwujnAknYXga4eA"
    REVIEW_ACTION_URL_FOR_INPATIENT: str = "https://forms.gle/ULRFv4aujQQsrvFG7"

    PERSON_VISIT_NOTIFICATION_TEXT_CANCEL_OR_REPLACE_RECEPTION: str = (
        "\nВ случае отмены или переноса записи обязательно свяжитесь с нами по номеру +7(423)2790790.\n С уважением, больница Пасифик Интернешнл Хоспитал."
    )

    PERSON_VISIT_NOTIFICATION_HEADER: str = (
        "_Здравствуйте, это *автоматическая* рассылка от Пасифик Интернешнл Хоспитал (Falck)._\n\n"
    )

    SEND_TELEGRAM_BOT_TEXT: str = (
        "\n\nОтправляем ссылку на наш telegram-бот с *важной информацией* (подготовка к исследованиям, врачи, услуги, схема проезда и др.):\n"
        + TELEGRAM_BOT_URL
    )

    ASK_TO_SEND_TELEGRAM_BOT_URL_TEXT: str = (
        "\n\nОтправьте в ответ любое сообщение и мы пришлём Вам ссылку на наш telegram-бот с *важной информацией* (подготовка к исследованиям, врачи, услуги, схема проезда и др.)"
    )

    PERSON_VISIT_NOTIFICATION_APPOINTMENT_INFORMATION: str = (
        "*{name}*, Вы записаны в Пасифик Интернешнл Хоспитал на {appointment_information}."
    )

    HAVE_A_GOOD_DAY: str = "\n\nХорошего дня!"

    PERSON_VISIT_GREETING_NOTIFICATION_TEXT_BASE: str = (
        PERSON_VISIT_NOTIFICATION_HEADER
        + PERSON_VISIT_NOTIFICATION_APPOINTMENT_INFORMATION
    )

    PERSON_VISIT_GREETING_NOTIFICATION_TEXT_WITHOUT_TEXT: str = (
        PERSON_VISIT_GREETING_NOTIFICATION_TEXT_BASE + ASK_TO_SEND_TELEGRAM_BOT_URL_TEXT
    )

    PERSON_VISIT_GREETING_NOTIFICATION_TEXT_FOR_CONFIRMED_NOTIFICATION: str = (
        PERSON_VISIT_GREETING_NOTIFICATION_TEXT_BASE
    )

    PERSON_VISIT_NOTIFICATION_WITH_TIME_TEXT: str = (
        "\n\nВаш приём запланирован на {day_string} {month_string} в {hour_string}{minute_string}."
        + PERSON_VISIT_NOTIFICATION_TEXT_CANCEL_OR_REPLACE_RECEPTION
    )

    PERSON_REVIEW_NOTIFICATION_TEXT_BASE: str = (
        "Добрый день, *{name}*!\n\nМеня зовут Анна, я директор отдела качества *Pacific International Hospital* (ранее Falck).\n\nВы недавно обращались в нашу больницу. Будем очень признательны, если в целях улучшения качества обслуживания вы ответите на несколько вопросов"
    )

    SEND_REVIEW_ACTION_URL_TEXT: str = ", перейдя по ссылке ниже:\n" + REVIEW_ACTION_URL

    PERSON_REVIEW_NOTIFICATION_TEXT_FOR_CONFIRMED_NOTIFICATION: str = (
        PERSON_REVIEW_NOTIFICATION_TEXT_BASE + SEND_REVIEW_ACTION_URL_TEXT
    )

    ASK_TO_SEND_REVIEW_ACTION_TEXT: str = ". Согласны ли Вы пройти опрос?"

    PERSON_REVIEW_NOTIFICATION_TEXT: str = (
        PERSON_REVIEW_NOTIFICATION_TEXT_BASE + ASK_TO_SEND_REVIEW_ACTION_TEXT
    )

    YES_ANSWER: tuple[str, ...] = (
        "да",
        "согласен",
        "согласна",
        "ok",
        "ок",
        "yes",
        "хорошо",
        "ага",
        "можно"
    )

    NO_ANSWER: tuple[str, ...] = ("нет", "не согласен", "не согласна", "no", "занят")

    TAKE_TELEGRAM_BOT_URL_TEXT: str = "*{name}*, отправляем Вам ссылку:\n"
    TAKE_TELEGRAM_BOT_URL_PERSONLESS_TEXT: str = "Отправляем Вам ссылку:\n"

    PERSONLESS_TAKE_TELEGRAM_BOT_URL_TEXT: str = "Отправляем Вам ссылку:\n"

    TAKE_REVIEW_ACTION_URL_TEXT: str = (
        "*{name}*, отправляем Вам ссылку для прохождения опроса:"
    )

    DATE_FORMAT: str = "%d.%m.%Y"
    DATE_IS_NOT_SET_YEAR: int = 1899
    DATETIME_FORMAT: str = "dd.mm.yyyy hh24:mi:ss"

    DB_DATETIME_FORMAT: str = "%d-%m-%Y-%H-%M-00"

    class BARCODE:

        class PERSON:
            IMAGE_FORMAT: str = FILE.EXTENSION.JPEG

        class PERSON_CARD_REGISTRY_FOLDER:
            IMAGE_FORMAT: str = FILE.EXTENSION.PNG

        NOT_FOUND: str = "_@barcode_not_found@_"
        ACTUAL_FORMAT: str = BARCODE.CODE128
        OLD_FORMAT: str = BARCODE.I25
        SUPPORT_FORMATS: list[str] = [ACTUAL_FORMAT, OLD_FORMAT]
        NEW_PREFIX: str = "new_"

        @staticmethod
        def get_file_name(pin: int, with_extension: bool = False) -> str:
            extension: str = (
                j((".", POLIBASE.BARCODE.PERSON.IMAGE_FORMAT)) if with_extension else ""
            )
            return j((POLIBASE.BARCODE.NEW_PREFIX, pin, extension))

    POLIBASE_PERSON_REVIEW_NOTIFICATION_DOCTOR_PERSON_PIN_LIST: list[int] = [
        51087,
        24727,
        8846,
        104498,
        97967,
        12411,
        40903,
        1121,
        100669,
        2455,
        99537,
        1226,
        67558,
        12097,
        43476,
        12146,
        22142,
        4497,
        9207,
        102114,
        34388,
        5960,
        111110,
        118252,
        120547,
        120508,
        122658,
    ]

    BONUS_DOCTOR_PERSON_PIN_LIST: list[int] = [
        51087,
        24727,
        8846,
        104498,
        97967,
        12411,
        40903,
        1121,
        1226,
        67558,
        12097,
        43476,
        12146,
        22142,
        4497,
        102114,
        34388,
        5960,
        111110,
        118252,
        120508,
    ]


class PolibasePersonVisitStatus:
    CONFIRMED: int = 107
    CANCELED: int = 102


class PolibasePersonVisitNotificationType(Enum):
    GREETING = auto()
    REMINDER = auto()
    DEFAULT = auto()


"""
102 - отмена		
99 прошу перенести
101 - пришел			
102 - отказался	
103 - на приеме
104 - окончен
105 - не пришел	
106 - предварительно
107 - подверждено
108 - оказано
109 к оплате
"""


class PolibaseDocumentTypes(Enum):

    ABPM_JOURNAL = PolibaseDocumentDescription(
        "Дневник суточного мониторинга АД", 70, 70, 80
    )
    HOLTER_JOURNAL = PolibaseDocumentDescription(
        "Дневник суточного мониторинга ЭКГ", 70, 70, 80
    )
    PATIENT_CARD_AMBULATORY = PolibaseDocumentDescription(
        "Медицинская карта\nпациента, получившего медицинскую помощь\nв амбулаторных условиях",
        70,
        120,
        120,
    )
    PROCESSING_PRESONAL_DATA_CONSENT = PolibaseDocumentDescription(
        "согласие\nпациента на обработку персональных данных",
        70,
        70,
        120,
        1,
    )
    INFORMED_VOLUNTARY_MEDICAL_INVENTION_CONSENT = PolibaseDocumentDescription(
        "информированное добровольное согласие\nна медицинское вмешательство",
        70,
        70,
        120,
        1,
    )
    INFORMED_VOLUNTARY_MEDICAL_INVENTION_CONSENT_SPECIFIC = PolibaseDocumentDescription(
        "информированное добровольное согласие на виды медицинских\nвмешательств, включенные в перечень определенных видов\nмедицинских вмешательств, на которые граждане дают\nинформированное добровольное согласие при выборе врача и\nмедицинской организации для получения первичной медико-\nсанитарной помощи",
        70,
        70,
        280,
        2,
    )

    @staticmethod
    def sorted() -> list:
        return sorted(
            PolibaseDocumentTypes,
            key=lambda item: item.value.title_height,
            reverse=True,
        )



import re
import os
import sys
import json
import shlex
import socket
import string
import random
import ntpath
import pathlib
import calendar
import platform
import dataclasses
from enum import Enum
from time import sleep
from os import walk, scandir
from operator import itemgetter
from contextlib import contextmanager
from inspect import isfunction, isclass
from typing import Any, Callable, Tuple, List
from datetime import date, datetime, timedelta

from pih.consts.password import RULES
from pih.consts.date_time import DATE_TIME
from pih.consts.names import FieldCollectionAliases

from pih.collections import (
    R,
    T,
    User,
    Result,
    FullName,
    FieldItem,
    FieldItemList,
    PolibasePerson,
)


def if_else(
    check_value: bool,
    true_value: Callable[[], Any] | Any,
    false_value: Callable[[], Any] | Any = None,
) -> Any:
    return (
        (true_value() if callable(true_value) else true_value)
        if check_value
        else (
            false_value()
            if not n(false_value) and callable(false_value)
            else false_value
        )
    )


def v(value: Any) -> Any:
    if n(value):
        if isinstance(value, tuple):
            return ()
        if isinstance(value, str):
            return ""
        if isinstance(value, list):
            return []
    return value


def one(
    value: Result[T] | Result[list[T]] | T | list[T], default_value: Any = None
) -> T | None:
    return (
        ResultTool.get_first_item(value, default_value)
        if isinstance(value, Result)
        else DataTool.get_first_item(value, default_value)
    )


def nl(
    value: str = "", count: int = 1, reversed: bool | None = False, normal: bool = True
) -> str:
    nl_text: str = ["<br>", "\n"][int(normal)] * count
    return DataTool.triple_bool(
        reversed, j((value, nl_text)), j((nl_text, value)), j((nl_text, value, nl_text))
    )


def j(value: tuple[Any, ...] | list[Any], splitter: str = "") -> str:
    return splitter.join(DataTool.map(str, DataTool.filter(nn, value)))

def nnl(value: Any) -> list:
    return value

def nna(value: Any) -> Any:
    return value

def nnd(value: Any) -> dict:
    return value

def nnt(value: T | None) -> T:
    return value

def nns(value: Any) -> str:
    return value

def nnb(value: Any) -> bytes:
    return value

def nni(value: Any) -> int:
    return value

def nndt(value: Any) -> datetime:
    return value

def nnf(value: Any) -> float:
    return value

def jnl(value: tuple[Any, ...] | list[Any]) -> str:
    return j(value, nl())


def jp(value: tuple[Any, ...] | list[Any]) -> str:
    return j(value, ".")


def js(value: tuple[Any, ...] | list[Any], aditional_splitter: str = "") -> str:
    return j(value, j((aditional_splitter or "", " ")))


class ParameterList:
    def __init__(self, value: Any):
        self.values = (
            value if isinstance(value, (list, tuple)) else [value]
        )
        self.index = 0

    def next(self, object: Any = None, default_value: Any = None) -> Any:
        if self.index >= len(self.values):
            return default_value
        value: Any = self.values[self.index]
        if value == "":
            value = None
        self.index = self.index + 1
        if ne(value) and ne(object):
            if isclass(object) and issubclass(object, Enum):
                value = EnumTool.get(object, value)
            else:
                value = DataTool.fill_data_from_source(object, value)
        return value

    def next_as_list(self, class_type) -> Any:
        return DataTool.fill_data_from_list_source(class_type, self.next())

    def get(
        self,
        index: int = 0,
        object: Any = None,
        default_value: Any = None,
    ) -> Any:
        temp_index: int = self.index
        self.index = index
        result: Any = self.next(object, default_value)
        self.index = temp_index
        return result

    def set(self, index: int, value: Any) -> None:
        self.values[index] = value


class EnumTool:
    @staticmethod
    def get(
        enum_class: Enum | Any,
        key: str | None = None,
        default_value: Any = None,
        return_value: bool = True,
    ) -> Any:
        if n(enum_class):
            return None
        if return_value and e(key):
            if isinstance(enum_class, Enum):
                return enum_class.value
            return enum_class
        if key not in enum_class._member_map_:
            return default_value
        return enum_class._member_map_[key]

    @staticmethod
    def get_by_value(enum_class: Enum, value: Any, default_value: Any = None) -> Any:
        if isinstance(value, Enum):
            return value
        map: Any = enum_class._value2member_map_
        return map[value] if value in map else default_value

    @staticmethod
    def get_by_value_or_key(
        enum_class: Enum, value: Any, default_value: Any = None
    ) -> Any:
        return EnumTool.get_by_value(enum_class, value, default_value) or EnumTool.get(
            enum_class, value, default_value
        )

    @staticmethod
    def get_value(enum_class: Enum | Any, default_value: str | None = None) -> Any:
        if nn(enum_class) and isinstance(enum_class, Enum):
            return enum_class.value
        return enum_class or default_value


class PIHEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, FieldItem):
            return {f"{obj.name}": obj.__dict__}
        if isinstance(obj, FieldItemList):
            return obj.list
        if isinstance(obj, Enum):
            return obj.name
        if isinstance(obj, ParameterList):
            return obj.values
        if dataclasses.is_dataclass(obj):
            return DataTool.to_data(obj)
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class DataTool:

    @staticmethod
    def init_var(name: str, holder: dict[str, Any], default_value: Any = None) -> None:
        if name not in holder:
            holder[name] = default_value

    @staticmethod
    def sort(sort_function: Callable, data: list[T], reverse: bool = False) -> list[T]:
        if nn(sort_function):
            data.sort(key=sort_function, reverse=reverse)
        return data

    @staticmethod
    def every(
        action_function: Callable[[T, int], None] | Callable[[T], None],
        data: list[T],
        use_index: bool = False,
    ) -> list[T]:
        if use_index:
            for index, item in enumerate(data):
                action_function(item, index)
        else:
            for item in data:
                action_function(item)
        return data

    @staticmethod
    def fields(value: object) -> list[str]:
        if dataclasses.is_dataclass(value):
            return [field.name for field in dataclasses.fields(value)]
        return []

    @staticmethod
    def map(
        function: Callable[[Any], Any], value: list[Any] | dict[Any, Any] | Any
    ) -> list[Any] | dict[Any, Any]:
        if isinstance(value, dict):
            return (
                {} if n(value) else {key: function(data) for key, data in value.items()}
            )
        return [] if n(value) else list(map(function, DataTool.as_list(value)))

    @staticmethod
    def filter(function: Callable[[Any], bool], value: list[Any] | Any) -> list[Any]:
        return [] if n(value) else list(filter(function, DataTool.as_list(value)))

    @staticmethod
    def as_value(
        function_or_value: Callable[[], str] | Callable[[Any], str] | str | None,
        parameters: Any = None,
    ) -> str:
        return (
            (function_or_value() if n(parameters) else function_or_value(parameters))
            if callable(function_or_value)
            else function_or_value
        ) or ""

    @staticmethod
    def as_bitmask_value(
        value: int | tuple[Enum, ...] | Enum | list[Enum] | list[int],
    ) -> int:
        value_list: list[Enum | int] | None = None
        if isinstance(value, (list, tuple)):
            value_list = value
        elif isinstance(value, (int, Enum)):
            value_list = [value]
        return BitMask.set(value_list)

    @staticmethod
    def by_index(data: list | None, index: int, default_value: Any = None) -> Any:
        if n(data):
            return default_value
        if len(data) <= index:
            return default_value
        return data[index]

    @staticmethod
    def rpc_encode(data: dict | None, ensure_ascii: bool = True) -> str | None:
        return (
            None
            if n(data)
            else json.dumps(data, cls=PIHEncoder, ensure_ascii=ensure_ascii)
        )

    @staticmethod
    def rpc_decode(value: str | None) -> dict[str, Any] | None:
        return None if e(value) else json.loads(value)

    @staticmethod
    def to_result(
        result_string: str | None,
        class_type_holder: Any | Callable[[Any], Any] | None = None,
        first_data_item: bool = False,
    ) -> Result[Any]:
        result_object: dict[str, Any] | None = DataTool.rpc_decode(result_string)
        if n(result_object):
            return Result(None, None)
        data: dict = result_object["data"]  # type: ignore
        data = DataTool.get_first_item(data) if first_data_item else data

        def fill_data_with(item: Any) -> Any:
            if e(class_type_holder):
                return item
            return (
                class_type_holder(item)
                if callable(class_type_holder) and not isclass(class_type_holder)
                else DataTool.fill_data_from_source(
                    (
                        class_type_holder()
                        if isclass(class_type_holder)
                        else class_type_holder
                    ),
                    item,
                )
            )

        def obtain_data() -> Any:
            return (
                list(map(fill_data_with, data))
                if isinstance(data, list)
                else fill_data_with(data)
            )

        if "fields_alias" in result_object:
            return Result(
                FieldItemList(
                    EnumTool.get(
                        FieldCollectionAliases, result_object["fields_alias"]
                    ).value
                ),
                obtain_data(),
            )
        else:
            fields = None if "fields" not in result_object else result_object["fields"]
        field_list: list[FieldItem] | None = None
        if nn(fields):
            field_list = []
            for field_item in fields:
                for field_name in field_item:
                    field_list.append(
                        DataTool.fill_data_from_source(
                            FieldItem(), field_item[field_name]
                        )
                    )
        return Result(FieldItemList(field_list) if field_list else None, obtain_data())

    @staticmethod
    def as_list(value: Any) -> list[Any]:
        if e(value):
            return []
        if isinstance(value, (list, tuple)):
            return value
        if isinstance(value, set):
            return list(value)
        if isinstance(value, dict):
            return list(value.values())
        try:
            if issubclass(value, Enum):
                return [EnumTool.get(item) for item in value]
        except TypeError:
            pass
        return [value]

    @staticmethod
    def to_list(value: dict | Enum, key_as_value: bool | None = False) -> list[Any]:
        if isinstance(value, dict):
            return [key if key_as_value else item for key, item in value.items()]
        result: list[Any | str] = []
        for item in value:
            result.append(
                [item.name, item.value]
                if n(key_as_value)
                else (item.name if key_as_value else item.value)
            )
        return result

    @staticmethod
    def triple_bool(
        value: bool | None, false_result: Any, true_result: Any, none_result: Any
    ) -> Any:
        if n(value):
            return none_result
        return true_result if value else false_result

    @staticmethod
    def to_result_with_fields(
        data: str, fields: FieldItemList, cls=None, first_data_item: bool = False
    ) -> Result:
        return Result(fields, DataTool.to_result(data, cls, first_data_item))

    @staticmethod
    def to_string(obj: object, join_symbol: str = "") -> str:
        return j(DataTool.to_list(DataTool.to_data(obj)), join_symbol)

    @staticmethod
    def to_data(obj: object) -> dict:
        return obj.__dict__

    @staticmethod
    def fill_data_from_source(
        destination: Any,
        source: Any,
        copy_by_index: bool = False,
        skip_not_none: bool = False,
    ) -> Any:
        if n(source):
            return None
        else:
            if dataclasses.is_dataclass(source):
                source = source.__dict__
            if copy_by_index:
                [
                    setattr(
                        destination, key.name, [source[key] for key in source][index]
                    )
                    for index, key in enumerate(dataclasses.fields(destination))
                ]
            else:
                if dataclasses.is_dataclass(source):
                    for field in destination.__dataclass_fields__:
                        if field in source:
                            if not skip_not_none or e(
                                destination.__getattribute__(field)
                            ):
                                destination.__setattr__(field, source[field])
                else:
                    is_dict: bool = isinstance(source, dict)
                    for field in destination.__dataclass_fields__:
                        if field in source if is_dict else hasattr(source, field):
                            if not skip_not_none or e(
                                destination.__getattribute__(field)
                            ):
                                destination.__setattr__(
                                    field,
                                    (
                                        source[field]
                                        if is_dict
                                        else source.__getattribute__(field)
                                    ),
                                )
        return destination

    @staticmethod
    def fill_data_from_list_source(
        class_type, source: list[Any] | dict[str, Any]
    ) -> Any:
        if n(source):
            return None
        return list(
            map(
                lambda item: DataTool.fill_data_from_source(class_type(), item),
                source if isinstance(source, list) else source.values(),
            )
        )

    @staticmethod
    def fill_data_from_rpc_str(data: T, source: str) -> T:
        return DataTool.fill_data_from_source(data, DataTool.rpc_decode(source))

    @staticmethod
    def get_first_item(
        value: list[T] | T | dict[str, Any], default_value: Any = None
    ) -> T | Any:
        if e(value):
            return default_value
        if isinstance(value, dict):
            for _, item in value.items():
                return item
        return value[0] if isinstance(value, (list, tuple)) else value

    @staticmethod
    def get_last_item(
        value: list[T] | T | dict[str, Any], default_value: Any = None
    ) -> T | Any:
        if e(value):
            return default_value
        if isinstance(value, dict):
            for _, item in reversed(value.items()):
                return item
        return value[len(value) - 1] if isinstance(value, (list, tuple)) else value

    @staticmethod
    def if_is_in(
        value: Any,
        arg_name: Any,
        default_value: Any | Callable[[], Any] | None = None,
    ) -> Any:
        return DataTool.check(
            DataTool.is_in(value, arg_name), lambda: value[arg_name], default_value
        )

    @staticmethod
    def is_in(value: Any, arg_name: Any) -> bool:
        if isinstance(value, (list, tuple)) and isinstance(arg_name, int):
            return arg_name < len(value)
        try:
            if issubclass(value, Enum):
                return arg_name in value.__members__
        except TypeError:
            pass
        return arg_name in value

    @staticmethod
    def check(
        check_value: bool,
        true_value: Callable[[], Any] | Any,
        false_value: Callable[[], Any] | Any = None,
    ) -> Any:
        return if_else(check_value, true_value, false_value)

    @staticmethod
    def check_not_none(
        check_value: Any | list[Any] | tuple[Any, ...] | None,
        true_value: Callable[[], Any] | Any,
        false_value: Callable[[], Any] | Any = None,
        check_all: bool = False,
    ) -> Any:
        check: bool = False
        if isinstance(check_value, (list, tuple)):
            for item in check_value:
                check = not n(item)
                if (not check_all and check) or (check_all and not check):
                    break
        else:
            check = not n(check_value)
        return (
            (true_value() if callable(true_value) else true_value)
            if check
            else (
                false_value()
                if not n(false_value) and callable(false_value)
                else false_value
            )
        )

    @staticmethod
    def if_not_empty(
        check_value: Any,
        return_value: Callable[[Any], Any],
        default_value: Any = None,
    ) -> Any:
        return default_value if e(check_value) else return_value(check_value)

    @staticmethod
    def is_empty(value: list | str | dict | tuple | Any) -> bool:
        return n(value) or (
            isinstance(value, (list, str, dict, tuple)) and len(value) == 0
        )

    @staticmethod
    def is_not_none(value: Any) -> bool:
        return not n(value)

    @staticmethod
    def is_none(value: Any) -> bool:
        return value is None


class FullNameTool:
    SPLIT_SYMBOL: str = " "
    FULL_NAME_LENGTH: int = 3

    @staticmethod
    def format(value: str) -> str:
        return FullNameTool.fullname_to_string(FullNameTool.fullname_from_string(value))

    @staticmethod
    def fullname_to_string(full_name: FullName, join_symbol: str = SPLIT_SYMBOL) -> str:
        return DataTool.to_string(full_name, join_symbol)

    @staticmethod
    def fullname_from_string(value: str, split_symbol: str = SPLIT_SYMBOL) -> FullName:
        full_name_string_list: list[str] = ListTool.not_empty_items(
            value.split(split_symbol)
        )
        return FullName(
            full_name_string_list[0], full_name_string_list[1], full_name_string_list[2]
        )

    @staticmethod
    def is_fullname(value: str, split_symbol: str = SPLIT_SYMBOL) -> bool:
        return (
            len(ListTool.not_empty_items(value.split(split_symbol)))
            >= FullNameTool.FULL_NAME_LENGTH
        )

    @staticmethod
    def is_equal(fn_a: FullName, fn_b: FullName) -> bool:
        return (
            fn_a.first_name == fn_b.first_name
            and fn_a.middle_name == fn_b.middle_name
            and fn_a.last_name == fn_b.last_name
        )

    @staticmethod
    def is_intersect(fn_a: FullName, fn_b: FullName) -> bool:
        al: list[str] = [fn_a.last_name, fn_a.first_name, fn_a.middle_name]
        bl: list[str] = [fn_b.last_name, fn_b.first_name, fn_b.middle_name]
        return (
            len([value for value in al if value in bl]) == FullNameTool.FULL_NAME_LENGTH
        )

    @staticmethod
    def to_given_name(
        full_name_holder: FullName | PolibasePerson | User | str,
        join_symbol: str = SPLIT_SYMBOL,
    ) -> str:
        if isinstance(full_name_holder, PolibasePerson):
            return FullNameTool.to_given_name(full_name_holder.FullName, join_symbol)
        if isinstance(full_name_holder, User):
            return FullNameTool.to_given_name(full_name_holder.name)
        if isinstance(full_name_holder, FullName):
            return join_symbol.join(
                ListTool.not_empty_items(
                    [full_name_holder.first_name, full_name_holder.middle_name]
                )
            )
        if isinstance(full_name_holder, str):
            full_name_holder = full_name_holder.strip()
            if FullNameTool.is_fullname(full_name_holder):
                return FullNameTool.to_given_name(
                    FullNameTool.fullname_from_string(full_name_holder, join_symbol)
                )
            else:
                return full_name_holder


class BitMask:
    @staticmethod
    def add(
        value: int | None, bit: int | tuple[Enum, ...] | Enum | list[Enum] | list[int]
    ) -> int:
        value = value or 0
        bits: list[int | Enum] = bit if isinstance(bit, (list, tuple)) else [bit]
        for bit in bits:
            if isinstance(bit, int):
                value |= bit
            elif isinstance(bit, Enum):
                value |= bit.value
        return value

    @staticmethod
    def set(bit: int | tuple[Enum, ...] | Enum | list[Enum] | list[int]) -> int:
        return BitMask.add(0, bit)

    @staticmethod
    def value(bit: int | tuple[Enum, ...] | Enum | list[Enum] | list[int]) -> int:
        return BitMask.add(0, bit)

    @staticmethod
    def has(
        value: int | tuple[Enum, ...] | Enum | list[Enum] | list[int],
        bit: int | tuple[Enum, ...] | Enum | list[Enum] | list[int],
    ) -> bool:
        if n(value):
            return False
        value = BitMask.value(value)
        bits: list[int] = bit if isinstance(bit, (list, tuple)) else [bit]
        result: bool = False
        if len(bits) > 1:
            for bit in bits:
                result = BitMask.has(value, bit)
                if result:
                    break
        else:
            if isinstance(bit, int):
                result = (value & bit) == bit
            elif isinstance(bit, Enum):
                result = BitMask.has(value, bit.value)
        return result

    @staticmethod
    def has_index(value: int, index: int) -> bool:
        return BitMask.has(value, pow(2, index))

    @staticmethod
    def remove(value: int, bit: int | Enum) -> int:
        if isinstance(bit, Enum):
            bit = bit.value
        if BitMask.has(value, bit):
            value ^= bit
        return value


class ResultTool:
    @staticmethod
    def pack(fields: Any, data: Any) -> dict[str, Any]:
        result: dict[str, Any] = {"data": data}
        if isinstance(fields, FieldCollectionAliases):
            result["fields_alias"] = fields.name
        else:
            result["fields"] = fields
        return result

    @staticmethod
    def unpack(result: dict) -> tuple[FieldItemList, Any]:
        return ResultTool.unpack_fields(result), ResultTool.unpack_data(result)

    @staticmethod
    def unpack_fields(data: dict) -> Any:
        if "fields_alias" in data:
            return (FieldCollectionAliases._member_map_[data["fields_alias"]].value,)
        return data["fields"]

    @staticmethod
    def unpack_data(result: dict) -> Any:
        return result["data"]

    @staticmethod
    def is_empty(result: Result | None) -> bool:
        return n(result) or e(result.data)

    @staticmethod
    def get_first_item(
        result: Result[list[T] | T], default_value: Any = None
    ) -> T | Any:
        return DataTool.get_first_item(result.data, default_value)

    @staticmethod
    def with_first_item(
        result: Result[list[T] | T], default_value: Any = None
    ) -> Result[T]:
        result.data = ResultTool.get_first_item(result, default_value)
        return result

    @staticmethod
    def to_string(
        result: Result[T],
        use_index: bool = True,
        item_separator: str = "\n",
        value_separator: str | None = None,
        show_caption: bool = True,
    ) -> str:
        result_string_list: list[str] = []
        data: list = DataTool.as_list(result.data)
        item_result_string_list: list[str] | None = None
        for index, data_item in enumerate(data):
            if use_index and len(data) > 1:
                result_string_list.append(f"*{str(index + 1)}*:")
            if nn(value_separator):
                item_result_string_list = []
            for field_item in result.fields.list:
                field: FieldItem = field_item
                if not field.visible:
                    continue
                data_value: str | None = None
                if isinstance(data_item, dict):
                    data_value = data_item[field.name]
                elif dataclasses.is_dataclass(data_item):
                    data_value = data_item.__getattribute__(field.name)
                data_value = data_value or "Нет"
                if n(value_separator):
                    if show_caption:
                        result_string_list.append(f"{field.caption}: {data_value}")
                    else:
                        result_string_list.append(data_value)
                else:
                    if show_caption:
                        item_result_string_list.append(f"{field.caption}: {data_value}")
                    else:
                        item_result_string_list.append(data_value)
            if nn(value_separator):
                result_string_list.append(value_separator.join(item_result_string_list))
        return item_separator.join(result_string_list)

    @staticmethod
    def as_list(result: Result[T]) -> Result[list[T]]:
        return Result(
            result.fields,
            (
                []
                if n(result.data)
                else [result.data] if not isinstance(result.data, list) else result.data
            ),
        )

    @staticmethod
    def filter(
        filter_function: Callable[[T], bool],
        result: Result[list[T]],
        as_new_result: bool = False,
    ) -> Result[list[T]]:
        try:
            data: list[T] = DataTool.filter(filter_function, result.data)
            if as_new_result:
                return Result(result.fields, data)
            result.data = data
        except StopIteration as _:
            pass
        return result

    @staticmethod
    def sort(
        sort_function: Callable, result: Result[list[T]], reverse: bool = False
    ) -> Result[list[T]]:
        if nn(sort_function):
            result.data.sort(key=sort_function, reverse=reverse)
        return result

    @staticmethod
    def every(
        action_function: Callable[[T, int], None] | Callable[[T], None],
        result: Result[list[T]],
        use_index: bool = False,
    ) -> Result[list[T]]:
        result.data = DataTool.every(action_function, result.data, use_index)
        return result

    @staticmethod
    def do_while(result: Result[list[T]], check_function: Callable[[T], bool]) -> Any:
        result_data: Any = None
        for item in result.data:
            if check_function(item):
                result_data = item
                break
        return result_data

    @staticmethod
    def map(
        map_function_or_class: Callable[[T], R] | R,
        result: Result[list[T]],
        map_on_each_data_item: bool = True,
        as_new_result: bool = False,
    ) -> Result[list[R]]:
        map_function: Callable[[T], R] | None = None
        if not isfunction(map_function_or_class):
            map_function = lambda item: DataTool.fill_data_from_source(
                map_function_or_class(), item
            )
        else:
            map_function = map_function_or_class
        data: list[R] = (
            list(map(map_function, result.data))
            if map_on_each_data_item
            else map_function_or_class(result.data)
        )
        if as_new_result:
            return Result(result.fields, data)
        else:
            result.data = data
        return result


n = DataTool.is_none
nn = DataTool.is_not_none


def lw(value: str | list[str] | tuple[str, ...] | None) -> str | list[str] | tuple[str, ...]:
    if n(value):
        return ""
    if isinstance(value, str):
        return value.lower()
    return DataTool.map(str.lower, value)


def e(value: Any | Result[Any]) -> bool:
    return (
        ResultTool.is_empty(value)
        if isinstance(value, Result)
        else DataTool.is_empty(value)
    )


ne = lambda item: not e(item)


def ln(value: Any) -> int:
    return 0 if n(value) else len(value)


def separate_on_parts(value: Any) -> tuple[str, str, str]:
    value = str(value)
    strip_value: str = value.strip()
    index: int = value.find(strip_value)
    left: str = value[0:index]
    right: str = value[index + len(strip_value) :]
    return (left, strip_value, right)


def i(value: Any) -> str:
    if ne(value):
        value = str(value).strip()
        if value.find("_") == -1:
            return j(("_", value, "_"))
    return value or ""


def b(value: Any) -> str:
    if ne(value):
        left, value, right = separate_on_parts(value)
        if value.find("*") == -1:
            return j((left, "*", value, "*", right))
    return value or ""


def lnk(value: str) -> str:
    return j(("{", value, "}"))


def esc(value: Any, single: bool = False) -> str:
    return f"'{value}'" if single else f'"{value}"'


def it(value: Any, action: Callable[[Any], None]) -> Any:
    action(value)
    return value


def escs(
    value: Any,
) -> str:
    return esc(value, True)


class ListTool:
    @staticmethod
    def diff(a: list[Any], b: list[Any]) -> list[Any]:
        return [i for i in a + b if i not in a or i not in b]

    @staticmethod
    def to_dict_with_none_value(
        value: list[Any], modificator: Callable[[Any], Any] | None = None
    ) -> dict[Any, None]:
        return {item if n(modificator) else modificator(item): None for item in value}  # type: ignore

    @staticmethod
    def to_dict(value: list[Any], key: str) -> dict[Any, Any]:
        return {getattr(item, key): item for item in value}

    @staticmethod
    def not_empty_items(value: list[Any]) -> list[Any]:
        return list(filter(lambda item: ne(item), value))

    @staticmethod
    def not_less_length_items(value: list[Any], length: int) -> list[Any]:
        return list(filter(lambda item: ne(item) and len(item) >= length, value))


class StringTool:
    @staticmethod
    def contains(value1: str, value2: str) -> bool:
        if n(value1) or n(value2):
            return False
        value1 = value1.lower()
        value2 = value2.lower()
        return (
            value2.find(value1) if len(value2) > len(value1) else value1.find(value2)
        ) != -1

    @staticmethod
    def equal(value1: str, value2: str) -> bool:
        if n(value1) or n(value2):
            return False
        value1 = value1.lower()
        value2 = value2.lower()
        return value1 == value2

    @staticmethod
    def split_with_not_empty_items(value: str, symbol: str = " ") -> list[str]:
        return ListTool.not_empty_items(value.split(symbol))

    @staticmethod
    def separate_unquoted_and_quoted(value: str) -> tuple[list[str], list[str]]:
        list_value: list[str] = shlex.split(value, posix=False)

        def is_quotes(value: str) -> bool:
            return value[0] == value[-1] and value[0] in ('"', "'")

        return (
            DataTool.filter(lambda item: not is_quotes(item), list_value),
            DataTool.map(
                lambda item: item[1:-1], DataTool.filter(is_quotes, list_value)
            ),
        )

    @staticmethod
    def list_to_string(
        value: list[str],
        escaped_string: bool = False,
        separator: str = ", ",
        start: str = "",
        end: str = "",
        filter_empty: bool = False,
    ) -> str:
        return (
            start
            + separator.join(
                list(
                    map(
                        lambda item: (
                            f"'{item}'"
                            if escaped_string
                            else str(item) if nn(item) else ""
                        ),
                        list(filter(lambda item: not filter_empty or ne(item), value)),
                    )
                )
            )
            + end
        )

    @staticmethod
    def capitalize(value: str) -> str:
        if e(value):
            return ""
        return value[0].upper() + ("" if len(value) == 1 else value[1:])

    @staticmethod
    def decapitalize(value: str) -> str:
        if e(value):
            return ""
        return value[0].lower() + ("" if len(value) == 1 else value[1:])

    @staticmethod
    def decapitalize(value: str) -> str:
        if e(value):
            return ""
        result: str = ""
        for index, char in enumerate(value):
            char_is_upper: bool = char.isupper()
            if char == " " or char_is_upper:
                char = char.lower() if char_is_upper else char
                result += char
                break
            result += char
        return result + value[index + 1 :] if index < len(value) else ""

    @staticmethod
    def from_russian_keyboard_layout(value: str) -> str:
        dictionary: dict[str, str] = {
            "Й": "Q",
            "Ц": "W",
            "У": "E",
            "К": "R",
            "Е": "T",
            "Н": "Y",
            "Г": "U",
            "Ш": "I",
            "Щ": "O",
            "З": "P",
            "Х": "{",
            "Ъ": "}",
            "Ф": "A",
            "Ы": "S",
            "В": "D",
            "А": "F",
            "П": "G",
            "Р": "H",
            "О": "J",
            "Л": "K",
            "Д": "L",
            "Ж": ":",
            "Э": '"',
            "Я": "Z",
            "Ч": "X",
            "С": "C",
            "М": "V",
            "И": "B",
            "Т": "N",
            "Ь": "M",
            "Б": "<",
            "Ю": ">",
            "Ё": "~",
            "й": "q",
            "ц": "w",
            "у": "e",
            "к": "r",
            "е": "t",
            "н": "y",
            "г": "u",
            "ш": "i",
            "щ": "o",
            "з": "p",
            "х": "[",
            "ъ": "]",
            "ф": "a",
            "ы": "s",
            "в": "d",
            "а": "f",
            "п": "g",
            "р": "h",
            "о": "j",
            "л": "k",
            "д": "l",
            "ж": ";",
            "э": "'",
            "я": "z",
            "ч": "x",
            "с": "c",
            "м": "v",
            "и": "b",
            "т": "n",
            "ь": "m",
            "б": ",",
            "ю": ".",
            "ё": "`",
        }
        result: str = ""
        for item in value:
            result += dictionary[item] if item in dictionary else item
        return result


class DateTimeTool:
    @staticmethod
    def day_count(date: date | datetime, month_count: int) -> int:
        result: int = 0
        year: int = date.year
        month: int = date.month
        current_month: int = month
        for _ in range(month_count):
            if current_month > 12:
                year += 1
                current_month = 1
            result += calendar.monthrange(year, current_month)[1]
            current_month += 1
        # return month_count*31
        return result

    @staticmethod
    def add_months(date: date | datetime, value: int) -> date | datetime:
        return date + timedelta(days=DateTimeTool.day_count(date, value))

    @staticmethod
    def seconds_to_days(value: int) -> float:
        return value / 60 / 60 / 24

    @staticmethod
    def yesterday() -> datetime:
        return DateTimeTool.today(-1, as_datetime=True)

    @staticmethod
    def begin_date(value: datetime | date | None = None) -> datetime | date:
        value = value or DateTimeTool.today(as_datetime=True)
        return (
            value.replace(hour=0, minute=0, second=0, microsecond=0)
            if isinstance(value, datetime)
            else value
        )

    @staticmethod
    def end_date(value: datetime | date | None = None) -> datetime | date:
        value = value or DateTimeTool.today(as_datetime=True)
        return (
            value.replace(hour=23, minute=59, second=59, microsecond=0)
            if isinstance(value, datetime)
            else value
        )

    @staticmethod
    def timestamp() -> int:
        return int(datetime.now().timestamp())

    @staticmethod
    def date_or_today_string(value: datetime | None, format: str | None = None) -> str:
        return (
            DateTimeTool.datetime_to_string(value, format)
            if nn(value)
            else DateTimeTool.today_string(format)
        )

    @staticmethod
    def today_string(format: str | None = None, delta_days: int = 0) -> str:
        return DateTimeTool.datetime_to_string(
            DateTimeTool.today(delta_days, as_datetime=True), format
        )

    @staticmethod
    def datetime_to_string(
        date: datetime | None, format: str | None = None
    ) -> str | None:
        if n(date):
            return None
        return DataTool.check_not_none(
            format, lambda: date.strftime(format), lambda: date.isoformat()
        )

    @staticmethod
    def date_to_string(date: datetime, format: str | None = None) -> str:
        return DateTimeTool.datetime_to_string(date.date(), format)

    @staticmethod
    def to_date_string(isoformat_date_string: str) -> str:
        return isoformat_date_string.split(DATE_TIME.SPLITTER)[0]

    @staticmethod
    def today(delta_days: int = 0, as_datetime: bool = False) -> date | datetime:
        value: date = (datetime.today() + timedelta(days=delta_days)).date()
        if as_datetime:
            return datetime.combine(value, datetime.min.time())
        return value

    @staticmethod
    def now(
        minute: int | None = None,
        second: int | None = None,
        use_microsecond: bool = False,
    ) -> datetime:
        result: datetime = datetime.now()
        if ne(minute):
            result = result.replace(minute=minute)
        if ne(second):
            result = result.replace(second=second)
        return result if use_microsecond else result.replace(microsecond=0)

    @staticmethod
    def now_to_string(format: str | None = None) -> str:
        return DateTimeTool.datetime_to_string(DateTimeTool.now(), format)

    @staticmethod
    def now_time_to_string(format: str | None = None, delta_minutes: int = 0) -> str:
        return DateTimeTool.datetime_to_string(
            DateTimeTool.now_time(delta_minutes), format
        )

    @staticmethod
    def now_time(delta_minutes: int = 0) -> datetime:
        return datetime.combine(date.today(), datetime.now().time()) + timedelta(
            minutes=delta_minutes
        )

    @staticmethod
    def datetime_from_string(
        value: str | None, format: str | None = None
    ) -> datetime | None:
        if e(value):
            return None
        return DataTool.check_not_none(
            format,
            lambda: datetime.strptime(value, format),
            lambda: datetime.fromisoformat(value),
        )

    @staticmethod
    def datetime_or_date_from_string(
        value: str | None, format: str | None = None, as_date: bool = False
    ) -> datetime | date | None:
        if e(value):
            return None
        result: datetime = DataTool.check_not_none(
            format,
            lambda: datetime.strptime(value, format),
            lambda: datetime.fromisoformat(value),
        )
        if value.find(DATE_TIME.SPLITTER) != -1 and not as_date:
            return result
        return result.date()

    @staticmethod
    def date_from_string(value: str) -> date | None:
        return DateTimeTool.datetime_or_date_from_string(value, as_date=True)

    @staticmethod
    def is_equal_by_time(date: datetime, value: tuple | list | datetime) -> bool | None:
        if isinstance(value, (tuple, list)):
            return date.hour == value[0] and date.minute == value[1]
        if isinstance(value, datetime):
            return date.hour == value.hour and date.minute == value.minute
        return None


class TranslateTool:
    @staticmethod
    def ru_to_en(value: str) -> str:
        return value.translate(
            dict(
                zip(
                    map(
                        ord,
                        "йцукенгшщзхъфывапролджэячсмитьбю.ё"
                        "ЙЦУКЕНГШЩЗХЪФЫВАПРОЛДЖЭЯЧСМИТЬБЮ,Ё",
                    ),
                    "qwertyuiop[]asdfghjkl;'zxcvbnm,./`"
                    'QWERTYUIOP{}ASDFGHJKL:"ZXCVBNM<>?~',
                )
            )
        )


class PathTool:
    
    @staticmethod
    def directory_size(path: str) -> int:
        result = 0
        dirs = [path]
        while dirs:
            next_dir = dirs.pop()
            with scandir(next_dir) as it:
                for entry in it:
                    if entry.is_dir(follow_symlinks=False):
                        dirs.append(entry.path)
                    else:
                        result += entry.stat(follow_symlinks=False).st_size
        return result

    
    @staticmethod
    def path(value: str) -> str:
        return value.replace("\\", "/")

    @staticmethod
    def for_windows(value: str) -> str:
        separator: str = "\\"
        value = value.replace("/", separator)
        if value.find(":") == -1 and not value.startswith(separator * 2):
            value = j((separator*2, value))
        return value

    @staticmethod
    def exists(path: str) -> bool:
        return os.path.exists(path)

    @staticmethod
    def convert_for_unix(path: str) -> str:
        prefix: str = os.sep * 2
        path = path.replace("\\", "/").replace("\\\\", "//")
        return if_else(path.startswith(prefix), "", prefix) + path

    @staticmethod
    def get_file_list(path: str, created_after: float | None = None) -> list[str]:
        file_list: list[tuple[str, float]] = [
            (file_path, int(os.path.getctime(file_path)))
            for file_path in [
                os.path.join(path, file_path)
                for file_path in [
                    file_path for file_path in next(walk(path), (None, None, []))[2]
                ]
            ]
        ]
        if nn(created_after):
            file_list = sorted(file_list, key=itemgetter(1), reverse=True)
            return [
                file_item[0] for file_item in file_list if file_item[1] > created_after
            ]
        return [file_item[0] for file_item in file_list]

    @staticmethod
    def make_directory_if_not_exists(path: str) -> bool | None:
        try:
            is_exist = os.path.exists(path)
            if not is_exist:
                os.makedirs(path)
                return True
            return False
        except:
            return None

    @staticmethod
    def get_current_full_path(file_name: str) -> str:
        return os.path.join(sys.path[0], file_name)

    @staticmethod
    def add_extension(file_path: str | None = None, extension: str | None = None) -> str:
        if nn(file_path):
            dot_index: int = file_path.find(".")
            if dot_index != -1:
                source_extension: str = file_path.split(".")[-1]
                if source_extension == extension:
                    file_path = file_path[0:dot_index]
        return jp((file_path, extension))

    @staticmethod
    def get_file_name(path: str, with_extension: bool = False):
        head, tail = ntpath.split(path)
        value = tail or ntpath.basename(head)
        if not with_extension:
            value = value[0 : value.rfind(".")]
        return value

    @staticmethod
    def get_file_directory(path: str):
        head, _ = ntpath.split(path)
        if head[-1] in ["\\", "/"]:
            head = head[:-1]
        return head

    @staticmethod
    def get_extension(
        file_path: str,
    ) -> str:
        dot_index: int = file_path.rfind(".")
        if dot_index != -1:
            return file_path[dot_index + 1 :].lower()
        return ""

    @staticmethod
    def replace_prohibited_symbols_from_path_with_symbol(
        path: str, replaced_symbol: str = "_"
    ) -> str:
        return (
            path.replace("\\", replaced_symbol)
            .replace("/", replaced_symbol)
            .replace("?", replaced_symbol)
            .replace("<", replaced_symbol)
            .replace(">", replaced_symbol)
            .replace("*", replaced_symbol)
            .replace(":", replaced_symbol)
            .replace('"', replaced_symbol)
        )

    @staticmethod
    def resolve(src_path: str, host_nane: str | None = None) -> str:
        src_path = str(pathlib.Path(src_path).resolve())
        if src_path[1] == ":" and nn(host_nane):
            lan_adress: str = f"\\\\{host_nane}\\"
            src_path = j((lan_adress, src_path[0], "$", src_path[2:]))
        return src_path


class NetworkTool:
    @staticmethod
    def next_free_port() -> int:
        with socket.socket() as soc:
            soc.bind(("", 0))
            return soc.getsockname()[1]


def while_not_do(
    check_action: Callable[[], bool | Any] | None = None,
    attemp_count: int | None = None,
    success_handler: Callable[[], None] | None = None,
    start_handler: Callable[[], None] | None = None,
    sleep_time: int | None = None,
    action: Callable[[], Any] | None = None,
    check_action_as_bool: bool = True,
) -> Any:
    result: Any = None
    while e(check_action) or (
        (not check_action()) if check_action_as_bool else n(result := check_action())
    ):
        if nn(start_handler):
            start_handler()
            start_handler = None
        if nn(action):
            action()
        if nn(attemp_count):
            if attemp_count == 0:
                break
            attemp_count -= 1
        if nn(sleep_time):
            sleep(sleep_time)
    if nn(success_handler):
        success_handler()
    return result


def while_excepted(
    check_action: Callable[[], Any],
    on_except: Callable[[Any], None] | None = None,
) -> Any:
    def action() -> Any:
        try:
            return check_action()
        except Exception as error:
            if nn(on_except):
                on_except(error)
            return None

    return while_not_do(action, check_action_as_bool=False)


class Clipboard:
    @staticmethod
    def copy(value: str):
        import pyperclip as pc

        pc.copy(value)


class UserTools:
    @staticmethod
    def get_given_name(user: User) -> str:
        return FullNameTool.to_given_name(user.name)


class PasswordTools:
    @staticmethod
    def check_password(value: str, length: int, special_characters: str) -> bool:
        regexp_string = (
            r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)(?=.*["
            + special_characters
            + r"])[A-Za-z\d"
            + special_characters
            + "]{"
            + str(length)
            + ",}$"
        )
        password_checker = re.compile(regexp_string)
        return nn(re.fullmatch(password_checker, value))

    @staticmethod
    def generate_random_password(
        length: int,
        special_characters: str,
        order_list: list[str],
        special_characters_count: int,
        alphabets_lowercase_count: int,
        alphabets_uppercase_count: int,
        digits_count: int,
        shuffled: bool,
    ):
        alphabets_lowercase = list(string.ascii_lowercase)
        alphabets_uppercase = list(string.ascii_uppercase)
        digits = list(string.digits)
        characters = list(string.ascii_letters + string.digits + special_characters)
        characters_count = (
            alphabets_lowercase_count
            + alphabets_uppercase_count
            + digits_count
            + special_characters_count
        )
        if characters_count > length:
            return
        password: list[str] = []
        for order_item in order_list:
            if order_item == RULES.SPECIAL_CHARACTER:
                for i in range(special_characters_count):
                    password.append(random.choice(special_characters))
            elif order_item == RULES.LOWERCASE_ALPHABET:
                for i in range(alphabets_lowercase_count):
                    password.append(random.choice(alphabets_lowercase))
            elif order_item == RULES.UPPERCASE_ALPHABET:
                for i in range(alphabets_uppercase_count):
                    password.append(random.choice(alphabets_uppercase))
            elif order_item == RULES.DIGIT:
                for i in range(digits_count):
                    password.append(random.choice(digits))
        if characters_count < length:
            random.shuffle(characters)
            for i in range(length - characters_count):
                password.append(random.choice(characters))
        if shuffled:
            random.shuffle(password)
        return j(password)


class OSTool:
    @staticmethod
    def host() -> str:
        return platform.node()

    @staticmethod
    def pid() -> int:
        return os.getppid()


class ErrorTool:
    @contextmanager
    def detect(
        final_action: Callable[[], None] | None = None,
        error_handler: Callable[[Any], None] | None = None,
    ):
        try:
            yield True
        except Exception as error:
            if nn(error_handler):
                error_handler(error)
        finally:
            if nn(final_action):
                final_action()



from pih.consts.hosts import Hosts
from pih.consts.addresses import ADDRESSES
from enum import Enum

class SSHHosts(Enum):
    EMAIL_SERVER = ADDRESSES.EMAIL_SERVER_ADDRESS
    SITE_API = ADDRESSES.API_SITE_ADDRESS
    SITE = ADDRESSES.SITE_ADDRESS
    SERVICES = Hosts.SERVICES.NAME


from enum import Enum
from pih.collections import Host


class Hosts(Enum):

    @property
    def NAME(self) -> str:
        return self.value.name
    
    @property
    def IP(self) -> str | None:
        return self.value.ip
    
    @property
    def ALIAS(self) -> str:
        return self.value.alias

    ORION = Host("orion", None, "192.168.100.90")

    SERVICES = Host("svshost", "zabbix", "192.168.100.95")

    BACKUP_WORKER = Host("backup_worker", "backup_worker", "192.168.100.11")

    WS255 = Host("ws-255", None, "192.168.100.138")

    WS816 = Host("ws-816", None, "192.168.254.81")

    WS735 = Host("ws-735", "shared_disk_owner", "192.168.254.102")

    DEVELOPER = WS735

    DC1 = Host("fmvdc1.fmv.lan", "dc1", "192.168.100.4")

    DC2 = Host("fmvdc2.fmv.lan", "dc2", "192.168.100.23")

    PRINTER_SERVER = DC1

    POLIBASE1 = Host(
        # shit - cause polibase is not accessable
        "polibase1.fmv.lan",
        "polibase1",
        "192.168.100.3"
    )

    POLIBASE2 = Host("polibase2.fmv.lan", "polibase2", "192.168.110.3")

    POLIBASE = POLIBASE1

    POLIBASE_TEST = POLIBASE2

    _1C = Host("fmv1c2", "1c", "192.168.100.22")

    NAS = Host("nas", "nas", "192.168.100.200")

    PACS_ARCHVE = Host("pacs_archive", "ea_archive", "192.168.110.108")



import ipih
from pih.collections.service import (
    Subscribtion,
    ServiceDescription,
    ServiceInformation,
    SubscribtionResult,
    SubscriberInformation,
    ServiceDescriptionBase,
    SubscribtionInformation,
)
from pih.consts.errors import Error
from pih.consts import SUBSCRIBTION_TYPE
import pih.rpc.rpcCommandCall_pb2 as pb2
from pih.consts.date_time import DATE_TIME
from pih.consts.rpc import RPC as CONST_RPC
from pih.tools import DataTool, OSTool, EnumTool
from pih.consts.service_roles import ServiceRoles
import pih.rpc.rpcCommandCall_pb2_grpc as pb2_grpc
from pih.consts.service_commands import ServiceCommands as SC
from pih.tools import ParameterList, BitMask as BM, j, n, nn, e, ne, nl
from pih.tools.service import ServiceRoleTool, ServiceTool, ServiceAdminTool

import sys
import grpc
import psutil
from grpc import Server
from threading import Thread
from typing import Any, Callable
from collections import defaultdict
from datetime import datetime, timedelta

STUB: ServiceDescription = ServiceDescription(name="Stub")


def GRPC_OPTIONS() -> list[tuple[str, int]]:
    size: int = 20 * 1024 * 1024
    return [
        ("grpc.max_receive_message_length", size),
        ("grpc.max_send_message_length", size),
    ]


class IService:
    def serve(
        self,
        service_role_or_description: ServiceRoles | ServiceDescriptionBase,
        call_handler: Callable[[str, ParameterList, Any], Any] | None = None,
        starts_handler: Callable[[Any], None] | Callable[[], None] | None = None,
        max_workers: int | None = None,
        #depends_on_list: list[ServiceRoles | ServiceDescription] | None = None,
        as_standalone: bool = False,
        show_output: bool = True,
    ) -> None:
        raise NotImplemented()

    def create_subscribtions(
        self, information: ServiceDescriptionBase | None = None
    ) -> None:
        raise NotImplemented()

    @property
    def subscribtions(self) -> list[SubscribtionInformation]:
        raise NotImplemented()

    def stop(self) -> Server:
        raise NotImplemented()

    def subscribe_on(
        self, sc: SC, type: int = SUBSCRIBTION_TYPE.ON_RESULT, name: str | None = None
    ) -> bool:
        raise NotImplemented()

    def unsubscribe(
        self, commnad_list: list[SC] | None = None, type: int | None = None
    ) -> bool:
        raise NotImplemented()

    def get_description(self) -> ServiceDescription:
        raise NotImplemented()

    def get_information(self) -> ServiceInformation:
        raise NotImplemented()


DEBUG: bool = False


class RPC:
    server: Server | None = None

    context: grpc.RpcContext

    class SESSION:
        start_time: datetime
        life_time: timedelta
        thread_map: list[int] = []

    @staticmethod
    def create_service() -> IService:
        return RPC.Service()

    @staticmethod
    def create_error(context, message: str = "", code: Any | None = None) -> Any:
        context.set_details(message)
        context.set_code(code)
        return pb2.rpcCommandResult()

    class UnaryService(pb2_grpc.UnaryServicer):
        def __init__(
            self,
            service: IService,
            description: ServiceDescription,
            call_handler: Callable[[str, ParameterList, Any], Any] | None = None,
            *args,
            **kwargs,
        ):
            self.service_information = description
            self.call_handler = call_handler
            self.service = service
            self.subscriber_information_map: dict[
                SC, dict[str, SubscriberInformation]
            ] = defaultdict(dict)

        def internal_call_handler(
            self, command_name: str, parameters: Any, context
        ) -> Any:
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A
            try:
                sc: SC | None = EnumTool.get(SC, command_name)
                parameter_list: ParameterList = ParameterList(parameters)
                if ne(sc):
                    if sc == SC.stop_service:
                        is_subscriber: bool = ServiceTool.is_service_as_listener(
                            self.service_information
                        )
                        if not is_subscriber:
                            RPC.call_service_command(
                                SC.on_service_stops, self.service_information
                            )
                        self.service.stop()
                        if not is_subscriber:
                            pid: int = OSTool.pid()
                            parent = psutil.Process(pid)
                            for child in parent.children(recursive=True):
                                child.kill()
                        return
                    if sc == SC.ping:
                        return DataTool.fill_data_from_source(
                            ServiceInformation(), self.service_information
                        )
                    if sc == SC.subscribe:
                        return self.subscribe_with_parameter_list(parameter_list)
                    if sc == SC.unsubscribe:
                        return self.unsubscribe_with_parameter_list(parameter_list)
                    if sc == SC.create_subscribtions:
                        self.service.create_subscribtions(
                            parameter_list.next(ServiceDescriptionBase())
                        )
                        return True
                    if sc == SC.update_service_information:
                        ServiceAdminTool.update_service_information(
                            parameter_list.next_as_list(ServiceInformation),
                            parameter_list.next(),
                        )
                        return True
                    if sc == SC.heart_beat:
                        date_string: str = A.D_Ex.parameter_list(parameter_list).get()
                        date: datetime = datetime.strptime(
                            date_string, DATE_TIME.ISO_DATETIME_FORMAT
                        )
                        parameter_list.set(0, date)
                        RPC.SESSION.life_time = date - RPC.SESSION.start_time
                return (
                    None
                    if e(self.call_handler)
                    else self.call_subscribers_after(
                        sc,
                        parameter_list,
                        self.call_handler(
                            command_name,
                            self.call_subscribers_before(sc, parameter_list),
                            context,
                        ),
                    )
                )
            except Exception as exception:
                A.ER.global_except_hook(exception)

        def unsubscribe_all(self) -> None:
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A

            def unsubscribe_all_internal() -> None:
                for sc in self.subscriber_information_map:
                    for service_information_name in self.subscriber_information_map[sc]:
                        subscriber_information: SubscriberInformation = (
                            self.subscriber_information_map[sc][
                                service_information_name
                            ]
                        )
                        DataTool.rpc_decode(
                            RPC.call_service(
                                subscriber_information.service_information,
                                SC.unsubscribe,
                                (sc, subscriber_information.name),
                            )
                        )
                self.subscriber_information_map = defaultdict(dict)

            Thread(target=unsubscribe_all_internal).start()

        def unsubscribe_with_parameter_list(self, value: ParameterList) -> None:
            description: ServiceDescription = value.next(ServiceDescription())
            role_description_name: str = description.name
            for service_command in self.subscriber_information_map:
                if (
                    role_description_name
                    in self.subscriber_information_map[service_command]
                ):
                    del self.subscriber_information_map[service_command][
                        role_description_name
                    ]

        def subscribe_with_parameter_list(self, value: ParameterList) -> bool:
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A
            service_information: ServiceDescriptionBase = value.next(
                ServiceDescriptionBase()
            )
            subscribtion_information: SubscribtionInformation = value.next(
                SubscribtionInformation()
            )
            subscribtion_information.service_command = EnumTool.get(
                SC, subscribtion_information.service_command
            )
            return self.subscribe(service_information, subscribtion_information)

        def subscribe(
            self,
            service_description: ServiceDescription,
            subscribtion_information: SubscribtionInformation,
        ) -> bool:
            sc: SC = subscribtion_information.service_command
            name: str | None = subscribtion_information.name
            type: int = subscribtion_information.type
            if sc in self.subscriber_information_map:
                if service_description.name in self.subscriber_information_map[sc]:
                    subscriber_information: SubscriberInformation = (
                        self.subscriber_information_map[sc][service_description.name]
                    )
                    if (
                        subscriber_information.service_information
                        == service_description
                        and BM.has(subscriber_information.type, type)
                    ):
                        self.subscriber_information_map[sc][
                            service_description.name
                        ] = SubscriberInformation(
                            type, name, service_information=service_description
                        )
                else:
                    self.subscriber_information_map[sc][service_description.name] = (
                        SubscriberInformation(
                            type, name, service_information=service_description
                        )
                    )
            else:
                self.subscriber_information_map[sc][service_description.name] = (
                    SubscriberInformation(
                        type, name, service_information=service_description
                    )
                )
            return True

        def call_subscribers_before(self, sc: SC, in_result: ParameterList):
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A
            out_result: ParameterList = in_result
            if sc in self.subscriber_information_map:
                for service_information_name in list(
                    self.subscriber_information_map[sc]
                ):
                    subscriber_information: SubscriberInformation = (
                        self.subscriber_information_map[sc][service_information_name]
                    )
                    service_information: ServiceInformation = (
                        subscriber_information.service_information
                    )
                    if (
                        BM.has(
                            subscriber_information.type, SUBSCRIBTION_TYPE.ON_PARAMETERS
                        )
                        and subscriber_information.enabled
                    ):
                        subscriber_information.available = RPC.check_availability(
                            service_information
                        )
                        if subscriber_information.available:
                            out_result = ParameterList(
                                DataTool.rpc_decode(
                                    RPC.call_service(
                                        service_information,
                                        sc,
                                        (
                                            in_result,
                                            SubscribtionResult(
                                                None,
                                                SUBSCRIBTION_TYPE.ON_PARAMETERS,
                                                True,
                                            ),
                                        ),
                                    )
                                )
                            )
                        else:
                            del self.subscriber_information_map[sc][
                                service_information_name
                            ]
            return out_result

        def call_subscribers_after(
            self, sc: SC, parameter_list: ParameterList, result: Any
        ) -> Any:
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A

            def internal_call_subscribers_after_sequentially(
                sc: SC,
                subscriber_list: list[SubscriberInformation],
                parameter_list: ParameterList,
                result: Any,
            ):
                for subscriber in subscriber_list:
                    role_description_item: ServiceDescription = (
                        subscriber.service_information
                    )
                    subscriber.available = RPC.check_availability(role_description_item)
                    if subscriber.available:
                        result = RPC.call_service(
                            role_description_item,
                            sc,
                            (
                                parameter_list,
                                SubscribtionResult(
                                    result,
                                    SUBSCRIBTION_TYPE.ON_RESULT_SEQUENTIALLY,
                                    True,
                                ),
                            ),
                        )
                    else:
                        if DataTool.if_is_in(
                            self.subscriber_information_map[sc],
                            role_description_item.name,
                        ):
                            del self.subscriber_information_map[sc][
                                role_description_item.name
                            ]

            def internal_call_subscribers_after(
                sc: SC,
                subscriber_information: SubscriberInformation,
                parameter_list: ParameterList,
                result: Any,
            ):
                service_information: ServiceDescriptionBase = (
                    subscriber_information.service_information
                )
                subscriber_information.available = RPC.check_availability(
                    service_information
                )
                if subscriber_information.available:
                    RPC.call_service(
                        service_information,
                        sc,
                        (
                            parameter_list,
                            SubscribtionResult(
                                result, SUBSCRIBTION_TYPE.ON_RESULT, True
                            ),
                        ),
                    )
                else:
                    if service_information.name in self.subscriber_information_map[sc]:
                        del self.subscriber_information_map[sc][
                            service_information.name
                        ]

            if sc in self.subscriber_information_map:
                after_sequentially_subscriber_information_list: list[
                    SubscriberInformation
                ] = []
                for servce_information_name in list(
                    self.subscriber_information_map[sc]
                ):
                    subscriber_information: SubscriberInformation = (
                        self.subscriber_information_map[sc][servce_information_name]
                    )
                    if subscriber_information.enabled:
                        if BM.has(
                            subscriber_information.type, SUBSCRIBTION_TYPE.ON_RESULT
                        ):
                            Thread(
                                target=internal_call_subscribers_after,
                                args=(
                                    sc,
                                    subscriber_information,
                                    parameter_list,
                                    result,
                                ),
                            ).start()
                        elif BM.has(
                            subscriber_information.type,
                            SUBSCRIBTION_TYPE.ON_RESULT_SEQUENTIALLY,
                        ):
                            after_sequentially_subscriber_information_list.append(
                                subscriber_information
                            )
                if ne(after_sequentially_subscriber_information_list):
                    Thread(
                        target=internal_call_subscribers_after_sequentially,
                        args=(
                            sc,
                            after_sequentially_subscriber_information_list,
                            parameter_list,
                            result,
                        ),
                    ).start()

            return result

        def rpcCallCommand(self, command, context):
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A
            parameters = command.parameters

            if ne(parameters):
                parameters = DataTool.rpc_decode(parameters)
            RPC.context = context
            result: Any = self.internal_call_handler(command.name, parameters, context)
            if n(context.code()):
                return pb2.rpcCommandResult(data=DataTool.rpc_encode(result)) # type: ignore
            return result

    class Service(IService):
        MAX_WORKERS: int = 10

        def __init__(self):
            self.description: ServiceDescription | None = None
            self.information: ServiceInformation | None = None
            self.subscribtions_map: dict[
                ServiceDescriptionBase, dict[SC, Subscribtion]
            ] = defaultdict(dict)
            self.server: Server | None = None

        def get_description(self) -> ServiceDescription:
            return self.description

        def get_information(self) -> ServiceInformation:
            return self.information

        def serve(
            self,
            service_role_or_description: ServiceRoles | ServiceDescriptionBase,
            call_handler: Callable[[str, ParameterList, Any], Any] | None = None,
            starts_handler: Callable[[IService], None] | Callable[[], None] | None = None,
            max_workers: int | None = None,
            #depends_on_list: list[ServiceRoles | ServiceDescription] | None = None,
            as_standalone: bool = False,
            show_output: bool = True,
        ) -> None:
            if DEBUG:
                from pih import A
                from pih import PIHThreadPoolExecutor
            else:
                A = sys.modules["pih.pih"].A
                PIHThreadPoolExecutor = sys.modules["pih.pih"].PIHThreadPoolExecutor
            max_workers = max_workers or RPC.Service.MAX_WORKERS
            service_description: ServiceDescription = (
                service_role_or_description
                if isinstance(service_role_or_description, ServiceDescriptionBase)
                else ServiceRoleTool.service_description(service_role_or_description)
            )  # type: ignore
            self.description = service_description
            with A.ER.detect():
                if not ServiceTool.is_service_as_listener(service_description):
                    A.SE.add_isolated_arg()
                    isolate_arg: str | None = None
                    try:
                        isolated_arg_value: str | None = A.SE.isolated_arg()
                        if nn(isolated_arg_value):
                            isolate_arg = str(isolated_arg_value).lower()
                    except AttributeError as error:
                        pass
                    if nn(isolate_arg):
                        if n(self.description.isolated):
                            self.description.isolated = isolate_arg not in [
                                "0",
                                "no",
                                "false",
                            ]
                        elif self.description.isolated:
                            self.description.isolated = isolate_arg not in [
                                "0",
                                "no",
                                "false",
                            ]
            # self.description.pih_version = self.description.pih_version or A.V.value
            self.description.host = ServiceAdminTool.create_host(self.description)
            self.description.port = ServiceAdminTool.create_port(self.description)
            self.information = DataTool.fill_data_from_source(
                ServiceInformation(standalone=as_standalone),
                self.description,
                skip_not_none=True,
            ) # type: ignore
            self.information.pid = OSTool.pid() # type: ignore
            if e(RPC.service):
                RPC.service_description = self.description
                RPC.service_information = self.information # type: ignore
                RPC.service = self
            if show_output:
                A.O.init()
                A.O.service_header(self.information) # type: ignore
                A.O.good("Service was started!")
            self.server = grpc.server(
                PIHThreadPoolExecutor(max_workers=max_workers),
                options=GRPC_OPTIONS(),
                # compression=grpc.Compression.Gzip,
            )
            if n(RPC.server):
                RPC.server = self.server
            pb2_grpc.add_UnaryServicer_to_server(
                RPC.UnaryService(self, self.description, call_handler), self.server
            )
            try:
                """
                if DEBUG:
                    from pih.tools import while_not_do
                else:
                    while_not_do = sys.modules["pih.pih"].while_not_do
                """
                self.server.add_insecure_port(
                    j((self.description.host, self.description.port), ":")
                )
                self.server.start()
                """
                if ne(depends_on_list):
                    A.O.write_line("Dependencies availability check...")
                    while_not_do(
                        lambda: len(
                            list(
                                filter(
                                    lambda item: A.SRV.check_on_accessibility(
                                        item
                                        if isinstance(item, ServiceDescription)
                                        else item.value
                                    ),
                                    depends_on_list,
                                )
                            )
                        )
                        == len(depends_on_list)
                    )
                    A.O.write_line("All dependencies are online")
                """
                RPC.SESSION.start_time = datetime.now().replace(second=0, microsecond=0)
                if nn(starts_handler):
                    starts_handler(self) # type: ignore
                self.information.subscribtions = self.subscribtions # type: ignore
                RPC.call_service_command(SC.on_service_starts, self.information)
                with A.ER.detect_interruption("Выход"):
                    self.server.wait_for_termination()
            except RuntimeError as error:
                A.E.service_was_not_started(self.description, j(error.args)) # type: ignore

        def create_subscribtions(
            self, information: ServiceDescriptionBase | None = None
        ) -> None:
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A

            def internal_create_subscribtion(
                subscribtions: dict[ServiceDescriptionBase, dict[SC, Subscribtion]],
                information: ServiceDescriptionBase,
            ):
                for sc in subscribtions[information]:
                    subscription: Subscribtion = subscribtions[information][sc]
                    RPC.call_service(
                        information,
                        SC.subscribe,
                        (
                            DataTool.fill_data_from_source(
                                ServiceDescriptionBase(), self.information
                            ),
                            DataTool.fill_data_from_source(
                                SubscribtionInformation(), subscription
                            ),
                        ),
                    )
                    subscription.available = True
                    subscription.enabled = True

            for description_item in (
                self.subscribtions_map if e(information) else [information]
            ):
                if A.SRV.check_on_accessibility(description_item):
                    internal_create_subscribtion(
                        self.subscribtions_map, description_item
                    )

        @property
        def subscribtions(self) -> list[Subscribtion]:
            result: list[Subscribtion] = []
            for description_item in self.subscribtions_map:
                for sc in self.subscribtions_map[description_item]:
                    result.append(self.subscribtions_map[description_item][sc])
            return result

        def subscribe_on(
            self,
            sc: SC,
            type: int = SUBSCRIBTION_TYPE.ON_RESULT,
            name: str | None = None,
        ) -> bool:
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A
            service_description: ServiceDescription = A.SRV.description_by_command(sc)
            if ne(service_description):
                if service_description != self.description:
                    subscribtion: Subscribtion | None = None
                    if sc not in self.subscribtions_map[service_description]:
                        subscribtion = Subscribtion(sc, type, name)
                        self.subscribtions_map[service_description][sc] = subscribtion
                    else:
                        subscribtion = self.subscribtions_map[service_description][sc]
                        subscribtion.type |= type
                    subscribtion.available = A.SRV.check_on_accessibility(
                        service_description
                    )
                    return subscribtion.available
            return False

        def unsubscribe(
            self, command_list: list[SC] | None = None, type: int | None = None
        ) -> None:
            if DEBUG:
                from pih import A
            else:
                A = sys.modules["pih.pih"].A
            for service_description in self.subscribtions_map:
                for sc in self.subscribtions_map[service_description]:
                    if sc in command_list:
                        DataTool.rpc_decode(
                            RPC.call_service(
                                service_description, SC.unsubscribe, (self.description,)
                            )
                        )

        def unsubscribe_all(
            self, service_role_or_description: ServiceRoles | ServiceDescriptionBase
        ) -> bool:
            pass
            # return self.unsubscribe(service_role_or_description)

        def stop(self) -> None:
            self.server.stop(0)

    class CommandClient:
        def __init__(self, host: str, port: int):
            self.stub = pb2_grpc.UnaryStub(
                grpc.insecure_channel(
                    j((host, port), ":"),
                    options=GRPC_OPTIONS(),
                    # compression=grpc.Compression.Gzip,
                )
            )

        def call_command(
            self, name: str, parameters: str | None = None, timeout: int | None = None
        ):
            return self.stub.rpcCallCommand(
                pb2.rpcCommand(name=name, parameters=parameters), timeout=timeout
            )

    @staticmethod
    def ping(
        service_role_or_description: ServiceRoles | ServiceDescriptionBase,
    ) -> ServiceInformation | None:
        try:
            service_information: ServiceDescriptionBase = (
                ServiceRoleTool.service_description(service_role_or_description)
            )
            return DataTool.fill_data_from_rpc_str(
                ServiceInformation(),
                RPC.call_service(
                    service_information,
                    SC.ping,
                    ((RPC.service_information or STUB).name),
                ),
            )
        except Error:
            return None

    @staticmethod
    def stop(
        service_role_or_description: ServiceRoles | ServiceDescriptionBase,
    ) -> ServiceInformation:
        return RPC.call_service(
            ServiceRoleTool.service_description(service_role_or_description),
            SC.stop_service,
        )

    @staticmethod
    def check_availability(
        service_role_or_description: ServiceRoles | ServiceDescriptionBase,
    ) -> bool:
        return ne(RPC.ping(service_role_or_description))

    @staticmethod
    def call_service(
        service_role_or_description: ServiceRoles | ServiceDescriptionBase | None,
        sc: SC,
        parameters: Any | None = None,
        timeout: int | None = None,
    ) -> str | None:
        if DEBUG:
            from pih import A
        else:
            A = sys.modules["pih.pih"].A
        service_description: ServiceDescriptionBase | None = (
            service_role_or_description
            if n(service_role_or_description)
            else ServiceRoleTool.service_description(service_role_or_description)
        )  # type: ignore
        try:
            if e(service_description):
                service_description = A.SRV.description_by_command(sc)
            service_host: str | None = A.SRV.get_host(service_description)  # type: ignore
            service_port: int | None = A.SRV.get_port(service_description)  # type: ignore
            if e(timeout):
                if (
                    e(RPC.service_information)
                    or RPC.service_information.isolated  # type: ignore
                    or service_description.isolated  # type: ignore
                ):
                    if sc == SC.ping:
                        timeout = CONST_RPC.TIMEOUT_FOR_PING
                    else:
                        timeout = CONST_RPC.TIMEOUT
            return (
                RPC.CommandClient(service_host, service_port)  # type: ignore
                .call_command(sc.name, DataTool.rpc_encode(parameters), timeout)
                .data
            )
        except grpc.RpcError as exception:
            code: tuple = exception.code()  # type: ignore
            details: str = j(
                (
                    "Service role name: ",
                    service_description.name,  # type: ignore
                    nl(),
                    "Host: ",
                    service_host,
                    ":",
                    service_port,
                    nl(),
                    "Command: ",
                    sc.name,
                    nl(),
                    "Details: ",
                    exception.details(),  # type: ignore
                    nl(),
                    "Code: ",
                    code,
                )
            )  # type: ignore
            A.O.init()
            A.ER.rpc_error_handler(service_description, exception, code, details, sc)  # type: ignore
            # raise Error(error.details(), code)

    @staticmethod
    def call_service_command(
        command: SC, parameters: Any | None = None, timeout: int | None = None
    ) -> str | None:
        return RPC.call_service(None, command, parameters, timeout)

    service: Service | None = None
    service_description: ServiceDescription | None = None
    service_information: ServiceInformation | None = None



from enum import Enum, auto


class ServiceCommands(Enum):
    ping = auto()
    subscribe = auto()
    unsubscribe = auto()
    create_subscribtions = auto()
    stop_service = auto()
    send_event = auto()
    serve_command = auto()
    # Service admin
    on_service_starts = auto()
    on_service_stops = auto()
    update_service_information = auto()
    get_service_information_list = auto()

    # Log
    send_log_message = auto()
    #
    send_message_to_user_or_workstation = auto()
    add_message_to_queue = auto()
    send_delayed_message = auto()
    # Documents
    create_user_document = auto()
    save_time_tracking_report = auto()
    save_xlsx = auto()
    create_barcodes_for_inventory = auto()
    create_barcode_for_polibase_person = auto()
    create_qr_code = auto()
    check_inventory_report = auto()
    get_inventory_report = auto()
    save_inventory_report_item = auto()
    close_inventory_report = auto()
    # Polibase
    get_polibase_person_by_pin = auto()
    get_polibase_persons_by_pin = auto()
    get_polibase_persons_by_card_registry_folder_name = auto()
    get_polibase_persons_by_name = auto()
    get_polibase_person_pin_list_with_old_format_barcode = auto()
    get_polibase_person_registrator_by_pin = auto()
    get_polibase_persons_pin_by_visit_date = auto()
    get_polibase_persons_by_telephone_number = auto()
    get_polibase_person_pin_by_login = auto()
    get_polibase_person_user_login_and_worstation_name_pair_list = auto()
    #
    get_polibase_person_visits_last_id = auto()
    search_polibase_person_visits = auto()
    #
    set_polibase_person_card_folder_name = auto()
    set_polibase_person_email = auto()
    set_barcode_for_polibase_person = auto()
    check_polibase_person_card_registry_folder_name = auto()

    # ActiveDirectory
    check_user_exists_by_login = auto()
    drop_user_cache = auto()
    drop_workstaion_cache = auto()
    get_user_by_full_name = auto()
    get_users_by_name = auto()
    # get_active_users_by_name = auto()
    get_user_by_login = auto()
    get_user_by_telephone_number = auto()
    get_template_users = auto()
    get_containers = auto()
    get_user_list_by_job_position = auto()
    get_user_list_by_group = auto()
    create_user_by_template = auto()
    set_user_telephone_number = auto()
    authenticate = auto()
    set_user_password = auto()
    set_user_status = auto()
    get_printer_list = auto()
    remove_user = auto()
    get_computer_description_list = auto()
    get_computer_list = auto()
    get_workstation_list_by_user_login = auto()
    get_user_by_workstation = auto()
    # Printer
    printers_report = auto()
    # Orion
    get_free_mark_list = auto()
    get_temporary_mark_list = auto()
    get_mark_person_division_list = auto()
    get_time_tracking = auto()
    get_mark_list = auto()
    get_owner_mark_for_temporary_mark = auto()
    get_mark_by_tab_number = auto()
    get_mark_by_person_name = auto()
    get_free_mark_group_statistics_list = auto()
    get_free_mark_list_by_group_id = auto()
    get_mark_list_by_division_id = auto()
    set_full_name_by_tab_number = auto()
    set_telephone_by_tab_number = auto()
    check_mark_free = auto()
    create_mark = auto()
    remove_mark_by_tab_number = auto()
    make_mark_as_free_by_tab_number = auto()
    make_mark_as_temporary = auto()
    # PolibaseDatabaseBackup
    create_polibase_database_backup = auto()
    # DataStorage::Settings
    set_settings_value = auto()
    get_settings_value = auto()
    # HeatBeat
    heart_beat = auto()
    # Notifier
    register_polibase_person_information_quest = auto()
    search_polibase_person_information_quests = auto()
    update_polibase_person_information_quest = auto()
    # Visit Cached
    update_polibase_person_visit_to_data_stogare = auto()
    search_polibase_person_visits_in_data_storage = auto()
    # Visit notification
    register_polibase_person_visit_notification = auto()
    search_polibase_person_visit_notifications = auto()
    # Notification confirmation
    search_polibase_person_notification_confirmation = auto()
    update_polibase_person_notification_confirmation = auto()
    #
    check_email_accessibility = auto()
    get_email_information = auto()
    #
    register_delayed_message = auto()
    search_delayed_messages = auto()
    update_delayed_message = auto()
    #
    execute_data_source_query = auto()
    # WORKSTATION
    reboot = auto()
    shutdown = auto()
    log_out = auto()
    # Robocopy::Job
    robocopy_start_job = auto()
    robocopy_get_job_status_list = auto()
    # DataStorage::Storage value
    set_storage_value = auto()
    get_storage_value = auto()
    # Resource Manager
    get_resource_status_list = auto()

    register_ct_indications_value = auto()
    get_last_ct_indications_value_container_list = auto()
    get_last_сhiller_indications_value_container_list = auto()
    #
    test = auto()
    #
    execute_ssh_command = auto()
    get_certificate_information = auto()
    get_unix_free_space_information_by_drive_name = auto()
    print_image = auto()
    #
    get_ogrn_value = auto()
    get_fms_unit_name = auto()
    #
    start_polibase_person_information_quest = auto()
    register_chiller_indications_value = auto()
    #
    add_gkeep_item = auto()
    get_gkeep_item_list_by_any = auto()
    # get_gkeep_item_list_by_name = auto()
    #
    create_note = auto()
    get_note = auto()
    get_note_list_by_label = auto()
    set_polibase_person_telephone_number = auto()
    kill_process = auto()
    #
    register_event = auto()
    get_event = auto()
    remove_event = auto()
    #
    get_polibase_person_operator_by_pin = auto()
    #
    get_barcode_list_information = auto()
    document_type_exists = auto()
    #
    listen_for_new_files = auto()
    #
    recognize_document = auto()
    #
    get_polibase_person_by_email = auto()
    #
    create_statistics_chart = auto()
    #
    execute_polibase_query = auto()
    #
    send_email = auto()
    #
    joke = auto()
    #
    printer_snmp_call = auto()
    #
    update_person_change_date = auto()
    #
    drop_note_cache = auto()
    #
    get_bonus_list = auto()
    #
    door_operation = auto()
    #
    mount_facade_for_linux_host = auto()
    #
    get_event_count = auto()
    #
    get_user_list_by_property = auto()


from enum import Enum
from pih.collections import FieldItem, FieldItemList, ParamItem


class LINK:
    
    ADMINISTRATOR_PASSWORD: str = "ADMINISTRATOR_PASSWORD"
    ADMINISTRATOR_LOGIN: str = "ADMINISTRATOR_LOGIN"

    DEVELOPER_LOGIN: str = "DEVELOPER_LOGIN"
    DEVELOPER_PASSWORD: str = "DEVELOPER_PASSWORD"

    USER_LOGIN: str = "USER_LOGIN"
    USER_PASSWORD: str = "USER_PASSWORD"

    SERVICES_USER_LOGIN: str = "SERVICES_LOGIN"
    SERVICES_USER_PASSWORD: str = "SERVICES_PASSWORD"

    DATABASE_ADMINISTRATOR_LOGIN: str = "DATABASE_ADMINISTRATOR_LOGIN"
    DATABASE_ADMINISTRATOR_PASSWORD: str = "DATABASE_ADMINISTRATOR_PASSWORD"


class USER_PROPERTIES:
    TELEPHONE_NUMBER: str = "telephoneNumber"
    EMAIL: str = "mail"
    DN: str = "distinguishedName"
    USER_ACCOUNT_CONTROL: str = "userAccountControl"
    LOGIN: str = "samAccountName"
    DESCRIPTION: str = "description"
    PASSWORD: str = "password"
    USER_STATUS: str = "userStatus"
    NAME: str = "name"


class FIELD_NAME_COLLECTION:
    HOST: str = "host"
    LIKE: str = "like"
    LAST_ID: str = "last_id"
    ACTION_NAME: str = "action_name"
    ACTION_DESCRIPTION: str = "action_description"
    FULL_NAME: str = "FullName"
    IP_ADDRESS: str = "ip_address"
    TYPE: str = "type"
    GROUP_NAME: str = "GroupName"
    GROUP_ID: str = "GroupID"
    COMMENT: str = "Comment"
    CARD_REGISTRY_FOLDER: str = "ChartFolder"
    DESTINATION: str = "destination"
    BIRTH: str = "Birth"
    TAB_NUMBER: str = "TabNumber"
    OWNER_TAB_NUMBER: str = "OwnerTabNumber"
    NAME: str = USER_PROPERTIES.NAME
    TITLE: str = "title"
    TEXT: str = "text"
    MIDNAME: str = "MidName"
    PERSON_ID: str = "pID"
    MARK_ID: str = "mID"
    ID: str = "id"
    PIN: str = "pin"
    PID: str = "pid"
    VISIT_ID: str = "visitID"
    MESSAGE_ID: str = "messageID"
    VALUE: str = "value"
    FILE: str = "file"
    DIVISION_NAME: str = "DivisionName"
    DIVISION_ID: str = "DivisionID"
    BARCODE: str = "barcode"
    EMAILED: str = "emailed"
    PROPERTIES: str = "properties"
    PARAMETERS: str = "parameters"
    MESSAGE: str = "message"
    STATUS: str = "status"
    FEEDBACK_CALL_STATUS: str = "feedbackCallStatus"
    REGISTRATION_DATE: str = "registrationDate"
    CABINET_ID: str = "cabinetID"
    DOCTOR_ID: str = "doctorID"
    DOCTOR_FULL_NAME: str = "doctorFullName"
    SERVICE_GROUP_ID: str = "serviceGroupID"
    PORT_NAME: str = "portName"
    TEMPERATURE: str = "temperature"
    HUMIDITY: str = "humidity"
    INDICATORS: str = "indicators"
    DATA: str = "data"
    COUNT: str = "count"
    FORCED: str = "forced"
    IMAGES: str = "images"

    SEARCH_ATTRIBUTE_LOGIN: str = USER_PROPERTIES.LOGIN
    SEARCH_ATTRIBUTE_NAME: str = USER_PROPERTIES.NAME

    TELEPHONE_NUMBER: str = USER_PROPERTIES.TELEPHONE_NUMBER
    EMAIL: str = f"e{USER_PROPERTIES.EMAIL}"
    DN: str = USER_PROPERTIES.DN
    LOGIN: str = USER_PROPERTIES.LOGIN
    ACTIVE_USERS_LOGIN: str = "active_" + USER_PROPERTIES.LOGIN
    DESCRIPTION: str = USER_PROPERTIES.DESCRIPTION
    PASSWORD: str = USER_PROPERTIES.PASSWORD
    ACCESSABLE: str = "accessable"
    STEP: str = "step"
    STEP_CONFIRMED: str = "stepConfirmed"
    GRADE: str = "grade"
    INFORMATION_WAY: str = "informationWay"
    TIME: str = "time"

    TIMESTAMP: str = "timestamp"
    DATE: str = "date"
    BEGIN_DATE: str = "beginDate"
    COMPLETE_DATE: str = "completeDate"
    RECIPIENT: str = "recipient"
    SENDER: str = "sender"
    ANSWER: str = "answer"
    STATE: str = "state"

    INVENTORY_NUMBER: str = "inventory_number"
    QUANTITY: str = "quantity"
    ROW: str = "row"
    NAME_COLUMN: str = "name_column"
    INVENTORY_NUMBER_COLUMN: str = "inventory_number_column"
    QUANTITY_COLUMN: str = "quantity_column"

    TEMPLATE_USER_CONTAINER: str = "templated_user"
    CONTAINER: str = "container"

    REMOVE: str = "remove"
    AS_FREE: str = "as_free"
    CANCEL: str = "cancel"

    WORKSTATION_NAME: str = "workstation_name"
    WORKSTATION_DESCRIPTION: str = "workstation_description"
    PERSON_NAME: str = "person_name"
    PERSON_PIN: str = "person_pin"
    REGISTRATOR_PERSON_NAME: str = "registrator_person_name"
    REGISTRATOR_PERSON_PIN: str = "registrator_person_pin"


class KEYWORD_COLLECTION:
    LIKE: str = FIELD_NAME_COLLECTION.LIKE


class FIELD_ITEM_COLLECTION:
    TAB_NUMBER: FieldItem = FieldItem(
        FIELD_NAME_COLLECTION.TAB_NUMBER, "Табельный номер"
    )
    OWNER_TAB_NUMBER: FieldItem = FieldItem(
        FIELD_NAME_COLLECTION.OWNER_TAB_NUMBER, "Табельный номер владельца"
    )
    FULL_NAME: FieldItem = FieldItem(FIELD_NAME_COLLECTION.FULL_NAME, "Полное имя")
    TEMPERATURE: FieldItem = FieldItem(
        FIELD_NAME_COLLECTION.TEMPERATURE, "Температура", data_formatter="{data}°C"
    )
    INDICATORS: FieldItem = FieldItem(
        FIELD_NAME_COLLECTION.INDICATORS,
        "Индикаторы",
        data_formatter="chiller_indications_value_indicators",
    )
    INDICATION_TIMESTAMP: FieldItem = FieldItem(
        FIELD_NAME_COLLECTION.TIMESTAMP,
        "Время снятия показаний",
        data_formatter="my_datetime",
    )


class FIELD_COLLECTION:
    INDEX: FieldItem = FieldItem("__Index__", "Индекс", True)
    POSITION: FieldItem = FieldItem(
        "position", "Расположение", True, default_value="Нет в реестре карт"
    )

    VALUE: FieldItem = FieldItem("", "Значение", True)
    VALUE_LIST: FieldItem = FieldItem("", "Список значений", True)

    class ORION:
        MARK_ACTION: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.REMOVE, "Удалить"),
            FieldItem(FIELD_NAME_COLLECTION.AS_FREE, "Сделать свободной"),
            FieldItem(FIELD_NAME_COLLECTION.CANCEL, "Оставить"),
        )

        GROUP_BASE: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.GROUP_NAME, "Группа доступа"),
            FieldItem(FIELD_NAME_COLLECTION.COMMENT, "Описание", False),
        )

        TAB_NUMBER_BASE: FieldItemList = FieldItemList(FIELD_ITEM_COLLECTION.TAB_NUMBER)

        FREE_MARK: FieldItemList = FieldItemList(TAB_NUMBER_BASE, GROUP_BASE)

        TAB_NUMBER: FieldItemList = FieldItemList(
            TAB_NUMBER_BASE,
            FieldItem(
                FIELD_NAME_COLLECTION.DIVISION_NAME,
                "Подразделение",
                default_value="Без подразделения",
            ),
            GROUP_BASE,
        ).position(FIELD_NAME_COLLECTION.DIVISION_NAME, 2)

        TEMPORARY_MARK: FieldItemList = FieldItemList(
            FIELD_ITEM_COLLECTION.TAB_NUMBER,
            FIELD_ITEM_COLLECTION.OWNER_TAB_NUMBER,
            FIELD_ITEM_COLLECTION.FULL_NAME,
            FieldItem(FIELD_NAME_COLLECTION.PERSON_ID, "Person ID", False),
            FieldItem(FIELD_NAME_COLLECTION.MARK_ID, "Mark ID", False),
        )

        PERSON: FieldItemList = (
            FieldItemList(
                TAB_NUMBER,
                FieldItem(FIELD_NAME_COLLECTION.TELEPHONE_NUMBER, "Телефон", True),
                FIELD_ITEM_COLLECTION.FULL_NAME,
            )
            .position(FIELD_NAME_COLLECTION.FULL_NAME, 1)
            .position(FIELD_NAME_COLLECTION.TELEPHONE_NUMBER, 2)
        )

        PERSON_DIVISION: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.ID, "ID", False),
            FieldItem(FIELD_NAME_COLLECTION.NAME, "Название подразделения"),
        )

        PERSON_EXTENDED: FieldItemList = FieldItemList(
            PERSON,
            FieldItem(FIELD_NAME_COLLECTION.PERSON_ID, "Person ID", False),
            FieldItem(FIELD_NAME_COLLECTION.MARK_ID, "Mark ID", False),
        )

        GROUP: FieldItemList = FieldItemList(
            GROUP_BASE, FieldItem(FIELD_NAME_COLLECTION.GROUP_ID, "Group id", False)
        ).visible(FIELD_NAME_COLLECTION.COMMENT, True)

        GROUP_STATISTICS: FieldItemList = FieldItemList(
            GROUP,
            FieldItem("Count", "Количество"),
        ).visible(FIELD_NAME_COLLECTION.COMMENT, False)

        TIME_TRACKING: FieldItemList = FieldItemList(
            FIELD_ITEM_COLLECTION.FULL_NAME,
            FIELD_ITEM_COLLECTION.TAB_NUMBER,
            FieldItem("TimeVal", "Время"),
            FieldItem("Remark", "Remark"),
            FieldItem("Mode", "Mode"),
        )

        TIME_TRACKING_RESULT: FieldItemList = FieldItemList(
            FIELD_ITEM_COLLECTION.FULL_NAME,
            FIELD_ITEM_COLLECTION.TAB_NUMBER,
            FieldItem("Date", "Дата"),
            FieldItem("EnterTime", "Время прихода"),
            FieldItem("ExitTime", "Время ухода"),
            FieldItem("Duration", "Продолжительность"),
        )

    class INRENTORY:
        ITEM: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.NAME, "Название инвентарного объекта"),
            FieldItem(FIELD_NAME_COLLECTION.INVENTORY_NUMBER, "Инвентарный номер"),
            FieldItem(FIELD_NAME_COLLECTION.QUANTITY, "Количество"),
            FieldItem(FIELD_NAME_COLLECTION.NAME_COLUMN, None, False),
            FieldItem(FIELD_NAME_COLLECTION.INVENTORY_NUMBER_COLUMN, None, False),
            FieldItem(FIELD_NAME_COLLECTION.QUANTITY_COLUMN, None, False),
        )

    class AD:
        COMPUTER_DESCRIPTION: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.NAME, "Имя компьютера"),
            FieldItem(FIELD_NAME_COLLECTION.DESCRIPTION, "Описание"),
            FieldItem(FIELD_NAME_COLLECTION.PROPERTIES, "Свойства", visible=False),
        )

        USER_ACTION: FieldItemList = FieldItemList(
            FieldItem(USER_PROPERTIES.TELEPHONE_NUMBER, "Изменить номер телефона"),
            FieldItem(USER_PROPERTIES.PASSWORD, "Изменить пароль"),
            FieldItem(USER_PROPERTIES.USER_STATUS, "Активировать или деактивировать"),
        )

        SERVER: FieldItemList = FieldItemList(
            COMPUTER_DESCRIPTION,
            FieldItem(FIELD_NAME_COLLECTION.ACCESSABLE, "Доступен"),
        )

        WORKSTATION: FieldItemList = FieldItemList(
            SERVER,
            FieldItem(FIELD_NAME_COLLECTION.LOGIN, "Имя залогированного пользователя"),
        )

        SEARCH_ATTRIBUTE: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.SEARCH_ATTRIBUTE_LOGIN, "Логин"),
            FieldItem(FIELD_NAME_COLLECTION.SEARCH_ATTRIBUTE_NAME, "Имя"),
        )

        CONTAINER: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.NAME, "Название"),
            FieldItem(FIELD_NAME_COLLECTION.DESCRIPTION, "Описание"),
        )

        USER_NAME: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.NAME, "Полное имя пользователя")
        )

        TEMPLATED_USER: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.DESCRIPTION, "Описание")
        )

        USER: FieldItemList = (
            FieldItemList(
                CONTAINER,
                FieldItem(FIELD_NAME_COLLECTION.LOGIN, "Логин"),
                FieldItem(FIELD_NAME_COLLECTION.TELEPHONE_NUMBER, "Телефон"),
                FieldItem(USER_PROPERTIES.EMAIL, "Электронная почта"),
                FieldItem(FIELD_NAME_COLLECTION.DN, "Размещение"),
                FieldItem("userAccountControl", "Свойства аккаунта", False),
            )
            .position(FIELD_NAME_COLLECTION.DESCRIPTION, 4)
            .caption(
                FIELD_NAME_COLLECTION.NAME,
                USER_NAME.get_item_by_name(FIELD_NAME_COLLECTION.NAME).caption,
            )
        )

        CONTAINER_TYPE: FieldItemList = FieldItemList(
            FieldItem(
                FIELD_NAME_COLLECTION.TEMPLATE_USER_CONTAINER, "Шаблонный пользователь"
            ),
            FieldItem(FIELD_NAME_COLLECTION.CONTAINER, "Контейнер"),
        )

    class POLIBASE:
        CARD_REGISTRY_FOLDER: FieldItem = FieldItem(
            FIELD_NAME_COLLECTION.CARD_REGISTRY_FOLDER,
            "Папка карты пациента",
            default_value="Не зарегистрирована в реестре карт пациентов",
        )

        PERSON_BASE: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.PIN, "Идентификационный номер пациента"),
            FieldItem(FIELD_NAME_COLLECTION.FULL_NAME, "ФИО пациента"),
            FieldItem(FIELD_NAME_COLLECTION.TELEPHONE_NUMBER, "Телефон"),
        )

        PERSON_VISIT: FieldItemList = FieldItemList(
            PERSON_BASE,
            FieldItem(FIELD_NAME_COLLECTION.REGISTRATION_DATE, "Дата регистрации"),
            FieldItem(FIELD_NAME_COLLECTION.DOCTOR_FULL_NAME, "Имя доктора"),
        )

        PERSON: FieldItemList = FieldItemList(
            PERSON_BASE,
            FieldItem(
                FIELD_NAME_COLLECTION.BIRTH,
                "День рождения",
                True,
                "datetime",
                data_formatter="my_date",
            ),
            FieldItem(
                FIELD_NAME_COLLECTION.EMAIL,
                "Электронная почта",
                default_value="Нет электронной почты",
            ),
            CARD_REGISTRY_FOLDER,
            FieldItem(FIELD_NAME_COLLECTION.COMMENT, "Комментарий"),
        )

    class POLICY:
        PASSWORD_TYPE: FieldItemList = FieldItemList(
            # FieldItem("EMAIL", "Для почты"),
            # FieldItem("SIMPLE", "Простой"),
            FieldItem("NORMAL", "Стандартный"),
            FieldItem("STRONG", "Сложный"),
        )

    class PRINTER:
        ITEM: FieldItemList = FieldItemList(
            FieldItem(FIELD_NAME_COLLECTION.NAME, "Name"),
            FieldItem("serverName", "Server name"),
            FieldItem(FIELD_NAME_COLLECTION.PORT_NAME, "Host name"),
            FieldItem(FIELD_NAME_COLLECTION.DESCRIPTION, "Description"),
            FieldItem("adminDescription", "Admin description", False),
            FieldItem("driverName", "Driver name"),
        )

    class INDICATIONS:
        CT_VALUE: FieldItemList = FieldItemList(
            FIELD_ITEM_COLLECTION.TEMPERATURE,
            FieldItem(
                FIELD_NAME_COLLECTION.HUMIDITY, "Влажность", data_formatter="{data}%"
            ),
        )

        CHILLER_VALUE: FieldItemList = FieldItemList(
            FIELD_ITEM_COLLECTION.TEMPERATURE, FIELD_ITEM_COLLECTION.INDICATORS
        )

        CT_VALUE_CONTAINER: FieldItemList = FieldItemList(
            CT_VALUE, FIELD_ITEM_COLLECTION.INDICATION_TIMESTAMP
        )

        CHILLER_VALUE_CONTAINER: FieldItemList = FieldItemList(
            CHILLER_VALUE, FIELD_ITEM_COLLECTION.INDICATION_TIMESTAMP
        )


class FieldCollectionAliases(Enum):
    TIME_TRACKING: FieldItem = FIELD_COLLECTION.ORION.TIME_TRACKING
    PERSON: FieldItem = FIELD_COLLECTION.ORION.PERSON
    TEMPORARY_MARK: FieldItem = FIELD_COLLECTION.ORION.TEMPORARY_MARK
    POLIBASE_PERSON: FieldItem = FIELD_COLLECTION.POLIBASE.PERSON
    POLIBASE_PERSON_VISIT: FieldItem = FIELD_COLLECTION.POLIBASE.PERSON_VISIT
    PERSON_DIVISION: FieldItem = FIELD_COLLECTION.ORION.PERSON_DIVISION
    PERSON_EXTENDED: FieldItem = FIELD_COLLECTION.ORION.PERSON_EXTENDED
    COMPUTER_DESCRIPTION: FieldItem = FIELD_COLLECTION.AD.COMPUTER_DESCRIPTION
    WORKSTATION: FieldItem = FIELD_COLLECTION.AD.WORKSTATION
    SERVER: FieldItem = FIELD_COLLECTION.AD.SERVER
    VALUE: FieldItem = FIELD_COLLECTION.VALUE
    VALUE_LIST: FieldItem = FIELD_COLLECTION.VALUE_LIST


class PARAM_ITEMS:
    HOST: ParamItem = ParamItem(FIELD_NAME_COLLECTION.HOST, "Хост")
    NAME: ParamItem = ParamItem(FIELD_NAME_COLLECTION.NAME, "")
    IP_ADDRESS: ParamItem = ParamItem(FIELD_NAME_COLLECTION.IP_ADDRESS, "")
    VALUE: ParamItem = ParamItem(FIELD_NAME_COLLECTION.VALUE, "Значение")
    TITLE: ParamItem = ParamItem(FIELD_NAME_COLLECTION.TITLE, "Заголовок")
    TEXT: ParamItem = ParamItem(FIELD_NAME_COLLECTION.TEXT, "Текст")
    PID: ParamItem = ParamItem(FIELD_NAME_COLLECTION.PID, "")
    PIN: ParamItem = ParamItem(FIELD_NAME_COLLECTION.PIN, "")
    STATUS: ParamItem = ParamItem(FIELD_NAME_COLLECTION.STATUS, "")
    PERSON_PIN: ParamItem = ParamItem(
        FIELD_NAME_COLLECTION.PERSON_PIN, "Идентификационный номер пациента"
    )
    PERSON_NAME: ParamItem = ParamItem(FIELD_NAME_COLLECTION.PERSON_NAME, "")
    REGISTRATOR_PERSON_NAME: ParamItem = ParamItem(
        FIELD_NAME_COLLECTION.REGISTRATOR_PERSON_NAME, ""
    )
    REGISTRATOR_PERSON_PIN: ParamItem = ParamItem(
        FIELD_NAME_COLLECTION.REGISTRATOR_PERSON_PIN, ""
    )
    CARD_REGISTRY_FOLDER: ParamItem = ParamItem(
        FIELD_NAME_COLLECTION.CARD_REGISTRY_FOLDER, ""
    )
    DESCRIPTION: ParamItem = ParamItem(FIELD_NAME_COLLECTION.DESCRIPTION, "")
    DESTINATION: ParamItem = ParamItem(FIELD_NAME_COLLECTION.DESTINATION, "")
    COUNT: ParamItem = ParamItem(FIELD_NAME_COLLECTION.COUNT, "Количество")
    TELEPHONE_NUMBER: ParamItem = ParamItem(
        FIELD_NAME_COLLECTION.TELEPHONE_NUMBER, "Телефонный номер"
    )
    TAB_NUMBER: ParamItem = ParamItem(
        FIELD_NAME_COLLECTION.TAB_NUMBER, "Табельный номер"
    )
    PASSWORD: ParamItem = ParamItem(FIELD_NAME_COLLECTION.PASSWORD, "")
    ID: ParamItem = ParamItem(FIELD_NAME_COLLECTION.ID, "Id")
    LOGIN: ParamItem = ParamItem(USER_PROPERTIES.LOGIN, "Login of user")
    FULL_NAME: ParamItem = ParamItem(FIELD_NAME_COLLECTION.FULL_NAME, "Name of user")
    TYPE: ParamItem = ParamItem(FIELD_NAME_COLLECTION.TYPE, "Type")
    SERVICE_NAME: ParamItem = ParamItem("service_name", "Name of service")
    PARAMETERS: ParamItem = ParamItem("parameters", "Parameters", optional=True)
    TAG: ParamItem = ParamItem("tag", "Tags")
    EMAIL: ParamItem = ParamItem(FIELD_NAME_COLLECTION.EMAIL, "")
    INPATIENT: ParamItem = ParamItem("inpatient", "")
    PATH: ParamItem = ParamItem("path", "Путь к файлу")
    DOCUMENT_NAME: ParamItem = ParamItem("document_name", "Имя документа")
    SECRET: ParamItem = ParamItem("secret", "Секретный код")
    LOCALY: ParamItem = ParamItem("localy", "Локально (на месте)")
    IMAGES: ParamItem = ParamItem("images", "Картинки")
    ARGUMENTS: ParamItem = ParamItem("arguments", "Аргументы")




from pih.tools import *
from pih.collections import *
from pih import A, PIH, OutputStub

from typing import Any

output_stub: OutputStub = OutputStub()


class Stub(PIH):
    def arg(self) -> Any | None:
        return self.session.arg()

    def __init__(self):
        self.is_force: bool = False


if not A.SE.is_mobile:
    self: Stub = Stub()
b = output_stub.b
i = output_stub.i


def execute(file_search_request: str) -> None:
    with A.ER.detect_interruption():
        A.R_F.execute(
            file_search_request,
            parameters={"self": self},
            stdout_redirect=False,
            catch_exceptions=True,
        )


def execute_file(
    use_authentification: bool,
    title: str | None,
    file_search_request: str,
    show_loading_delay: float | None = None,
    loading_text: str | None = None,
    done_text: str | None = None
) -> None:
    A.O.init()
    if not use_authentification or A.SE.authenticate():
        A.O.clear_screen()
        A.O.pih_title()
        if ne(title):
            A.O.head1(title)  # type: ignore
        if n(show_loading_delay):
            execute(file_search_request)
        else:
            with A.O.make_loading(
                show_loading_delay, loading_text or "Идёт выполнение файла"
            ):
                execute(file_search_request)
        A.O.good(done_text or "Выполнено!")
        A.SE.exit(0)
    A.SE.exit(0, "Выход")


def include(
    value: str, class_name: str | None = None, global_dict: dict[str, Any] | None = None
) -> Any | dict[str, Any] | None:
    dictionary: dict[str, Any] = A.R_F.execute(value, stdout_redirect=False)
    if n(class_name):
        return dictionary
    import_data: Any = dictionary[class_name]
    if n(global_dict):
        global_dict[class_name] = import_data
        return
    else:
        return import_data



import ipih

from pih import A

SR = A.CT_SR
SC = A.CT_SC

#version 1.0

ROLE: SR = SR.DEVELOPER
if A.U.for_service(ROLE):

    from typing import Any
    from pih.tools import ParameterList

    def service_call_handler(sc: SC, parameter_list: ParameterList) -> Any:     
        return None
    
    def service_starts_handler() -> None:
        A.SRV_A.subscribe_on(SC.print_image)
       
    A.SRV_A.serve(ROLE, service_call_handler, service_starts_handler)



EVENT_LISTENER_NAME_PREFIX: str = "_@@EventListener@@_"
SUPPORT_NAME_PREFIX: str = "_@@Support@@_"

SERVICE_DESCRIPTION_HOLDER_VARIABLE_NAME: str = "SD"

from enum import Enum
from pih.consts.hosts import Hosts
from pih.collections import Host
from pih.collections.service import ServiceDescription

class ServiceRoleBase(Enum):

    @property
    def value(self) -> ServiceDescription:
        return super().value

    @property
    def host(self) -> str | None:
        return self.value.host
    
    @property
    def standalone_name(self) -> str | None:
        return self.value.standalone_name

    @host.setter
    def host(self, value: str | Hosts | Host) -> None:
        if isinstance(value, str):
            self.value.host = value
        elif isinstance(value, Hosts):
            self.host = value.value
        elif isinstance(value, Host):
            self.host = value.name
            
    @property
    def NAME(self) -> str | None:
        return self.value.name


from dataclasses import dataclass, field
from datetime import datetime, timedelta, date
from enum import Enum
from typing import Any, Tuple, List, Callable, TypeVar, Generic
from collections import namedtuple


@dataclass
class Host:
    name: str
    aliases: str | tuple[str, ...] | None = None
    ip: str | None = None
    description: str | None = None

    @property
    def alias(self) -> str:
        if isinstance(self.aliases, str):
            return self.aliases
        return (
            self.aliases[0]
            if self.aliases is not None and len(self.aliases) > 0
            else self.name
        )


@dataclass
class ZabbixHost:
    id: str | None = None
    name: str | None = None
    host: str | None = None


@dataclass
class ZabbixMetricsValue:
    value: Any | None = None
    clock: datetime | int | None = None


@dataclass
class ZabbixMetrics:
    itemid: int | None = None
    type: int = 0
    snmp_oid: str | None = None
    hostid: int = 0
    name: str | None = None
    key_: str | None = None
    delay: str | None = None
    history: str | None = None
    trends: str | None = None
    status: int = 0
    value_type: int = 0
    trapper_hosts: tuple[str, ...] | None = None
    units: str | None = None
    formula: str | None = None
    logtimefmt: str | None = None
    templateid: int = 0
    valuemapid: int = 0
    params: tuple[str, ...] | None = None
    ipmi_sensor: str | None = None
    flags: int = 0
    interfaceid: int = 0
    description: str | None = None
    inventory_link: int = 0
    lifetime: str | None = None
    evaltype: int = 0
    jmx_endpoint: str | None = None
    master_itemid: int = 0
    timeout: str | None = None
    url: str | None = None
    query_fields: tuple[str, ...] | None = None
    posts: str | None = None
    status_codes: int = 0
    follow_redirects: int = 0
    lastclock: datetime | None = None
    lastvalue: Any = None
    prevvalue: Any = None


@dataclass
class PasswordSettings:
    length: int
    special_characters: str
    order_list: list[str]
    special_characters_count: int
    alphabets_lowercase_count: int
    alphabets_uppercase_count: int
    digits_count: int = 1
    shuffled: bool = False


@dataclass
class NameCaption:
    name: str | None
    caption: str | None = None


@dataclass
class NameCaptionDescription(NameCaption):
    description: str | None = None


@dataclass
class OrderedNameCaptionDescription(NameCaptionDescription):
    order: int | None = None


@dataclass
class IconedOrderedNameCaptionDescription(OrderedNameCaptionDescription):
    icon: str | None = None


@dataclass
class ParamItem(NameCaptionDescription):
    optional: bool = False

    visible: bool = True
    saved: bool = True
    key: bool = False

    def configurate(
        self,
        visible: bool | None = None,
        saved: bool | None = None,
        key: bool | None = None,
    ) -> Any:
        result: ParamItem = ParamItem(
            self.name, self.caption, self.description, self.optional
        )
        if visible is not None:
            result.visible = visible
        if saved is not None:
            result.saved = saved
        if key is not None:
            result.key = key
        return result


@dataclass
class FieldItem:
    name: str | None = None
    caption: str | None = None
    visible: bool = True
    class_type: Any | None = None
    default_value: str | None = None
    data_formatter: str = "{data}"


class FieldItemList:
    list: list[FieldItem]

    def copy_field_item(self, value: FieldItem) -> FieldItem:
        return FieldItem(
            value.name,
            value.caption,
            value.visible,
            value.class_type,
            value.default_value,
            value.data_formatter,
        )

    def __init__(self, *args):
        self.list = []
        arg_list = list(args)
        for arg_item in arg_list:
            if isinstance(arg_item, FieldItem):
                item: FieldItem = self.copy_field_item(arg_item)
                self.list.append(item)
            elif isinstance(arg_item, FieldItemList):
                for item in arg_item.list:
                    self.list.append(self.copy_field_item(item))
            elif isinstance(arg_item, list):
                self.list.extend(arg_item)

    def get_list(self) -> list[FieldItem]:
        return self.list

    def get_item_and_index_by_name(self, value: str) -> Tuple[FieldItem, int]:
        index: int = -1
        result: FieldItem | None = None
        for item in self.list:
            index += 1
            if item.name == value:
                result = item
                break
        return result, -1 if result is None else index

    def get_item_by_name(self, value: str) -> FieldItem:
        result, _ = self.get_item_and_index_by_name(value)
        return result

    def position(self, name: str, position: int):
        _, index = self.get_item_and_index_by_name(name)
        if index != -1:
            self.list.insert(position, self.list.pop(index))
        return self

    def get_name_list(self):
        return list(map(lambda item: str(item.name), self.list))

    def get_caption_list(self):
        return list(
            map(lambda x: str(x.caption), filter(lambda y: y.visible, self.list))
        )

    def visible(self, name: str, value: bool):
        item, _ = self.get_item_and_index_by_name(name)
        if item is not None:
            item.visible = value
        return self

    def caption(self, name: str, value: bool):
        item, _ = self.get_item_and_index_by_name(name)
        if item is not None:
            item.caption = value
        return self

    def length(self) -> int:
        return len(self.list)


T = TypeVar("T")
R = TypeVar("R")


@dataclass
class Result(Generic[T]):
    fields: FieldItemList | None = None
    data: T | None = None

    def __len__(self):
        return len(self.data)

    def __iadd__(self, value):
        if (
            isinstance(value, Result)
            and isinstance(self.data, list)
            and isinstance(value.data, list)
        ):
            self.data += value.data
        return self

    def __add__(self, value):
        if (
            isinstance(value, Result)
            and isinstance(self.data, list)
            and isinstance(value.data, list)
        ):
            self.data += value.data
        return self


@dataclass
class FullName:
    last_name: str = ""
    first_name: str = ""
    middle_name: str = ""

    def as_list(self) -> list[str]:
        return [self.last_name, self.first_name, self.middle_name]


@dataclass
class UserBase:
    name: str | None = None
    description: str | None = None
    distinguishedName: str | None = None


@dataclass
class User(UserBase):
    samAccountName: str | None = None
    mail: str | None = None
    telephoneNumber: str | None = None
    userAccountControl: int | None = None


Rect = namedtuple("Rect", ["left", "top", "width", "height"])


@dataclass
class IndicationDevice:
    name: str | None = None
    description: str | None = None
    ip_address: tuple[str, int] | None = None


@dataclass
class MailboxInfo:
    timestamp: datetime | None = None
    last_uid: str | None = None


@dataclass
class NewMailMessage:
    mailbox_address: str | None = None
    subject: str | None = None
    text: str | None = None
    from_: str | None = None


@dataclass
class RecipientWaitingForInput:
    group_name: str | None = None
    timeout: int | None = None
    recipient: str | None = None
    timestamp: datetime | None = None


@dataclass
class BarcodeInformation:
    data: str | None = None
    type: str | None = None
    rect: Rect | None = None


@dataclass
class EmailInformation:
    email: str | None = None
    person_pin: int = 0
    person_name: str | None = None


@dataclass
class InaccesableEmailInformation(EmailInformation):
    workstation_name: str | None = None
    registrator_person_name: str | None = None


@dataclass
class CardRegistryFolderPosition:
    ChartFolder: str | None = None
    p_a: int = 0
    p_b: int = 0
    p_c: int = 0


@dataclass
class ActionValue:
    caption: str
    value: str


@dataclass
class LoginPasswordPair:
    login: str | None = None
    password: str | None = None


@dataclass
class OGRN:
    name: str | None = None
    code: str | None = None
    data: dict | None = None


@dataclass
class ComputerDescription:
    name: str | None = None
    properties: int = 0
    description: str | None = None


@dataclass
class Computer(ComputerDescription):
    samAccountName: str | None = None
    accessable: bool | None = None


@dataclass
class Server(Computer):
    pass


@dataclass
class Workstation(Computer):
    pass


@dataclass
class ResourceDescription:
    address: str | None = None
    name: str | None = None
    inaccessibility_check_values: tuple[int, ...] = (2, 20, 15)


@dataclass
class ResourceDescriptionDelegated(ResourceDescription):
    delegator: str | None = None


@dataclass
class SiteResourceDescription(ResourceDescription):
    check_certificate_status: bool = False
    check_free_space_status: bool = False
    driver_name: str | None = None
    internal: bool = False


class IResourceStatus:
    pass


@dataclass
class ResourceStatus(ResourceDescription, IResourceStatus):
    accessable: bool | None = None
    inaccessibility_counter: int = 0
    inaccessibility_counter_total: int = 0


@dataclass
class WSResourceStatus(ResourceStatus):
    pass


@dataclass
class ServerResourceStatus(ResourceStatus):
    pass


@dataclass
class DiskStatistics:
    name: str | None = None
    free_space: int | None = None
    size: int | None = None


@dataclass
class DisksStatisticsStatus(IResourceStatus):
    host: str | None = None
    disk_list: list[DiskStatistics] = field(default_factory=list)


@dataclass
class SiteResourceStatus(ResourceStatus, SiteResourceDescription):
    certificate_status: str | None = None
    free_space_status: str | None = None


@dataclass
class MarkPerson:
    FullName: str | None = None
    TabNumber: str | None = None


@dataclass
class MarkPersonDivision(MarkPerson):
    DivisionName: str | None = None
    DivisionID: int | None = None


@dataclass
class TemporaryMark(MarkPerson):
    OwnerTabNumber: str | None = None


@dataclass
class PolibasePersonBase:
    pin: int | None = None
    FullName: str | None = None
    telephoneNumber: str | None = None


@dataclass
class PolibasePerson(PolibasePersonBase):
    Birth: datetime | None = None
    Comment: str | None = None
    ChartFolder: str | None = None
    email: str | None = None
    barcode: str | None = None
    registrationDate: datetime | None = None
    telephoneNumber2: str | None = None
    telephoneNumber3: str | None = None
    telephoneNumber4: str | None = None


@dataclass
class PolibaseNote:
    emailed: str | None = None


@dataclass
class PolibasePersonVisitDS(PolibasePersonBase):
    id: int | None = None
    registrationDate: str | None = None
    beginDate: str | datetime | None = None
    completeDate: str | datetime | None = None
    status: int | None = None
    cabinetID: int | None = None
    doctorID: int | None = None
    doctorFullName: str | None = None
    serviceGroupID: int | None = None


@dataclass
class CardRegistryFolderStatistics:
    name: str | None = None
    count: int = 0


@dataclass
class PolibasePersonVisitSearchCritery:
    vis_no: Any | None = None
    vis_pat_no: Any | None = None
    vis_pat_name: Any | None = None
    vis_place: Any | None = None
    vis_reg_date: Any | None = None
    vis_date_ps: Any | None = None
    vis_date_pf: Any | None = None
    vis_date_fs: Any | None = None
    vis_date_ff: Any | None = None


@dataclass
class PolibasePersonVisitNotificationDS:
    visitID: int | None = None
    messageID: int | None = None
    type: int | None = None


@dataclass
class EventDS:
    name: str | None = None
    parameters: dict | None = None
    timestamp: datetime | date | str | int | None = None
    id: int = 0


@dataclass
class Message:
    message: str | None = None
    recipient: str | None = None
    sender: str | None = None


@dataclass
class DelayedMessage(Message):
    date: Any | None = None
    type: int | None = None


@dataclass
class DelayedMessageDS(DelayedMessage):
    id: int | None = None
    status: int | None = None


@dataclass
class MessageSearchCritery:
    id: Any | None = None
    recipient: str | None = None
    date: datetime | str | None = None
    type: Any | None = None
    status: int | None = None
    sender: str | None = None


@dataclass
class PolibasePersonNotificationConfirmation:
    recipient: str | None = None
    sender: str | None = None
    status: int = 0


@dataclass
class PolibasePersonVisitNotification(
    PolibasePersonVisitDS, PolibasePersonVisitNotificationDS
):
    pass


@dataclass
class PolibasePersonVisit(PolibasePersonVisitDS):
    registrationDate: datetime | None = None
    beginDate: datetime | None = None
    completeDate: datetime | None = None
    beginDate2: datetime | None = None
    completeDate2: datetime | None = None


@dataclass
class PolibasePersonQuest:
    step: int | None = None
    stepConfirmed: bool | None = None
    timestamp: int | None = None


@dataclass
class PolibasePersonInformationQuest(PolibasePersonBase):
    confirmed: int | None = None
    errors: int | None = None


@dataclass
class PolibasePersonReviewQuest(PolibasePersonQuest):
    beginDate: str | None = None
    completeDate: str | None = None
    grade: int | None = None
    message: str | None = None
    informationWay: int | None = None
    feedbackCallStatus: int | None = None


@dataclass
class MarkGroup:
    GroupName: str | None = None
    GroupID: int | None = None


@dataclass
class Mark(MarkPersonDivision, MarkGroup):
    pID: int | None = None
    mID: int | None = None
    Comment: str | None = None
    telephoneNumber: str | None = None
    type: int | None = None


@dataclass
class PersonDivision:
    id: int | None = None
    name: str | None = None


@dataclass
class TimeTrackingEntity(MarkPersonDivision):
    TimeVal: str | None = None
    Mode: int | None = None


@dataclass
class TimeTrackingResultByDate:
    date: str | None = None
    enter_time: str | None = None
    exit_time: str | None = None
    duration: int | None = None


@dataclass
class TimeTrackingResultByPerson:
    tab_number: str | None = None
    full_name: str | None = None
    duration: int = 0
    list: List[TimeTrackingResultByDate] = field(default_factory=list)


@dataclass
class WhatsAppMessage:
    message: str | None = None
    from_me: bool | None = None
    sender: str | None = None
    recipient: str | None = None
    profile_id: str | None = None
    time: int | None = None
    chatId: str | None = None
    flags: int | None = None
    return_result_key: str | None = None
    args: tuple[Any, ...] | None = None


@dataclass
class WhatsAppMessagePayload:
    title: str
    body: str


@dataclass
class WhatsAppMessageListPayload(WhatsAppMessagePayload):
    btn_text: str
    list: dict


@dataclass
class WhatsAppMessagebButton:
    body: str | None = None
    id: str | None = None


@dataclass
class WhatsAppMessageButtonsPayload(WhatsAppMessagePayload):
    buttons: list[WhatsAppMessagebButton] | None = None


@dataclass
class TimeTrackingResultByDivision:
    name: str
    list: List[TimeTrackingResultByPerson] = field(default_factory=list)


@dataclass
class RobocopyJobDescription:
    name: str | None = None
    start_datetime: str | None = None
    host: str | None = None
    run_from_system_account: bool = False
    run_with_elevetion: bool = False
    live: bool = False
    exclude: bool = False
    
    def clone(
        self,
        job_name: str,
        start_datetime: str | None = None,
        host: str | None = None,
        live: bool | None = None,
        exclude: bool = False
    ):
        return RobocopyJobDescription(
            job_name,
            start_datetime,
            host or self.host,
            self.run_from_system_account,
            self.run_with_elevetion,
            self.live if live is None else live,
            self.exclude if exclude is None else exclude,
        )


@dataclass
class RobocopyJobItem(RobocopyJobDescription):
    source: str | None = None
    destination: str | None = None


@dataclass
class RobocopyJobStatus:
    name: str | None = None
    source: str | None = None
    destination: str | None = None
    active: bool = False
    last_created: str | None = None
    last_status: int | None = None
    live: bool = False
    pid: int = -1
    exclude: bool = False


@dataclass
class PrinterADInformation:
    driverName: str | None = None
    adminDescription: str | None = None
    description: str | None = None
    portName: str | None = None
    serverName: str | None = None
    name: str | None = None


@dataclass
class IndicationsContainer:
    timestamp: datetime | None = None


@dataclass
class HumidityIndicationsValue:
    humidity: float | None = None


@dataclass
class TemparatureIndicationsValue:
    temperature: float | None = None


@dataclass
class TemperatureAndHumidityIndicationsValue(
    HumidityIndicationsValue, TemparatureIndicationsValue
):
    pass


@dataclass
class ChillerIndicationsValue(TemparatureIndicationsValue):
    indicators: int = 0


@dataclass
class ChillerIndicationsValueContainer(ChillerIndicationsValue, IndicationsContainer):
    pass


@dataclass
class CTIndicationsValue(TemperatureAndHumidityIndicationsValue):
    pass


@dataclass
class CTIndicationsValueContainer(CTIndicationsValue, IndicationsContainer):
    pass


@dataclass
class GKeepItem:
    name: str | None = None
    title: str | None = None
    id: str | None = None


@dataclass
class File:
    title: str | None = None
    text: str | None = None
    id: str | None = None


@dataclass
class Note(File):
    images: list[str] | None = None


@dataclass
class InventoryReportItem:
    name: str | int | None = None
    inventory_number: str | None = None
    row: str | None = None
    quantity: int | None = None
    name_column: int | None = None
    inventory_number_column: int | None = None
    quantity_column: int | None = None


@dataclass
class PrinterStatus:
    ip: str | None = None
    description: str | None = None
    variant: str | None = None
    port: int | None = None
    community: str | None = None
    accessable: bool | None = None


@dataclass
class TimeSeriesStatistics:
    count: int = 0
    values: list[datetime] | None = None
    distance: list[timedelta] | None = None
    min: int | None = None
    max: int | None = None
    avg: int | None = None


@dataclass
class PrinterReport(PrinterStatus):
    name: str | None = None
    model: str | None = None
    serial: int | None = None
    adminDescription: str | None = None
    meta: str | None = None
    printsOverall: int | None = None
    printsColor: int | None = None
    printsMonochrome: int | None = None
    fuserType: int | None = None
    fuserCapacity: int | None = None
    fuserRemaining: int | None = None
    wasteType: int | None = None
    wasteCapacity: int | None = None
    wasteRemaining: int | None = None
    cleanerType: int | None = None
    cleanerCapacity: int | None = None
    cleanerRemaining: int | None = None
    transferType: int | None = None
    transferCapacity: int | None = None
    transferRemaining: int | None = None
    blackTonerType: str | None = None
    blackTonerCapacity: int | None = None
    blackTonerRemaining: int | None = None
    cyanTonerType: int | None = None
    cyanTonerCapacity: int | None = None
    cyanTonerRemaining: int | None = None
    magentaTonerType: int | None = None
    magentaTonerCapacity: int | None = None
    magentaTonerRemaining: int | None = None
    yellowTonerType: int | None = None
    yellowTonerCapacity: int | None = None
    yellowTonerRemaining: int | None = None
    blackDrumType: str | None = None
    blackDrumCapacity: int | None = None
    blackDrumRemaining: int | None = None
    cyanDrumType: int | None = None
    cyanDrumCapacity: int | None = None
    cyanDrumRemaining: int | None = None
    magentaDrumType: int | None = None
    magentaDrumCapacity: int | None = None
    magentaDrumRemaining: int | None = None
    yellowDrumType: int | None = None
    yellowDrumCapacity: int | None = None
    yellowDrumRemaining: int | None = None

    def get_toner(self, color: str) -> int:
        color = color.lower()
        remaining: int
        capacity: int
        if color == "c":
            remaining = self.cyanTonerRemaining
            capacity = self.cyanTonerCapacity
        if color == "m":
            remaining = self.magentaTonerRemaining
            capacity = self.magentaTonerCapacity
        if color == "y":
            remaining = self.yellowTonerRemaining
            capacity = self.yellowTonerCapacity
        if color == "k":
            remaining = self.blackTonerRemaining
            capacity = self.blackTonerCapacity
        try:
            if remaining == -1 or capacity == -1:
                return -1
            if remaining == -404 or capacity == -404:
                return -404
            if remaining == -401 or capacity == -401:
                return -401
            return int(round((int(remaining) / int(capacity)) * 100))
        except:
            return -1

    def get_drum(self, color: str) -> int:
        color = color.lower()
        if color == "c":
            remaining = self.cyanDrumRemaining
            capacity = self.cyanDrumCapacity
        if color == "m":
            remaining = self.magentaDrumRemaining
            capacity = self.magentaDrumCapacity
        if color == "y":
            remaining = self.yellowDrumRemaining
            capacity = self.yellowDrumCapacity
        if color == "k":
            remaining = self.blackDrumRemaining
            capacity = self.blackDrumCapacity
        try:
            if remaining == -1 or capacity == -1:
                return -1
            if remaining == -404 or capacity == -404:
                return -404
            if remaining == -401 or capacity == -401:
                return -401
            return int(round((int(remaining) / int(capacity)) * 100))
        except:
            return -1


@dataclass
class MarkGroupStatistics(MarkGroup):
    Comment: str | None = None
    Count: int | None = None


@dataclass
class EventDescription:
    message: str | Callable[[tuple[Any, ...] | list[Any]], str] | None = None
    channel: Enum | None = None
    flags: int | tuple[Enum, ...] | Enum | None = None
    params: tuple[ParamItem, ...] | None = None


@dataclass
class ActionDescription:
    name: str | None = None
    alias: tuple[str, ...] | None = None
    description: str | None = None
    question: str | None = None
    confirm: bool = True
    silence: bool = False
    parameters_description: str | None = None
    # parameters_default: tuple[str] | None = None
    forcable: bool = False
    forced_description: str | None = None


@dataclass
class ActionWasDone:
    action_description: str | None = None
    action: Enum | str | None = None
    user_name: str | None = None
    user_login: str | None = None
    parameters: list[Any] | None = None
    forced: bool = False


@dataclass
class StorageVariableHolder:
    key_name: str | None = None
    default_value: str | None = None
    description: str | None = None
    auto_init: bool = True
    # only for get
    section: str | None = None


@dataclass
class ExpiredTimestampVariableHolder:
    timestamp: str | None = None
    life_time: str | None = None
    resolver_note: str | None = None


@dataclass
class IntStorageVariableHolder(StorageVariableHolder):
    default_value: int = 0


@dataclass
class MinIntStorageVariableHolder(IntStorageVariableHolder):
    min_value: int = 0


@dataclass
class IntStorageVariableHolderWithMin(IntStorageVariableHolder):
    min_value: int = 0


@dataclass
class IntListStorageVariableHolder(StorageVariableHolder):
    default_value: list[int] | None = None


@dataclass
class FloatStorageVariableHolder(StorageVariableHolder):
    default_value: float | None = None


@dataclass
class BoolStorageVariableHolder(StorageVariableHolder):
    default_value: bool | None = None


@dataclass
class TimeStorageVariableHolder(StorageVariableHolder):
    default_value: str | None = None


@dataclass
class DateListStorageVariableHolder(StorageVariableHolder):
    default_value: tuple[str, ...] | None = None


@dataclass
class StringListStorageVariableHolder(StorageVariableHolder):
    default_value: tuple[str, ...] | None = None


@dataclass
class PolibaseDocument:
    file_path: str | None = None
    polibase_person_pin: int | None = None
    document_type: str | None = None


@dataclass
class MedicalDirectionDocument:
    number: int | None = None
    date: datetime | None = None
    person_name: str | None = None
    person_ensurence_number: str | None = None
    person_ensurence_agent: str | None = None
    person_birthday: datetime | str | None = None
    research_type: Enum | str | None = None
    research_code: str | None = None
    ogrn_number: str | None = None


@dataclass
class Titled:
    title: str


@dataclass
class ThresholdedText(Titled):
    threshold: float


@dataclass
class PolibaseDocumentDescription(ThresholdedText):
    title_top: int = 0
    title_height: int = 0
    page_count: int = 1


@dataclass
class DocumentDescription(ThresholdedText):
    left: float
    top: float
    right: float
    bottom: float


@dataclass
class MedicalResearchType:
    title_list: tuple[str, ...] | None = None
    alias: str | None = None


@dataclass
class DirectoryInfo:
    value: str | None = None
    confirmed_file_list: list[tuple[str, float]] | None = None
    last_created_file_timestamp: float | None = None


@dataclass
class JournalRecord:
    # from pih.const import JournalType, Tags
    timestamp: datetime | None = None
    applicant_user: User | None = None
    type: Any | None = None
    tag: Any | None = None
    title: str | None = None
    text: str | None = None
    parameters: dict[str, Any] | None = None


@dataclass
class BonusInformation:
    bonus_all: float = 0
    money_all: float = 0
    bonus_spent_all: float = 0
    bonus_active: float = 0
    money_last: float = 0
    bonus_last_spent: float = 0
    bonus_last: float = 0


@dataclass
class WappiStatus:

    app_status: str | None = None
    authorized: bool = False
    authorized_at: str | None = None
    checked_at: str | None = None
    last_activity: float = 0
    logouted_at: str | None = None
    message_count: int = 0
    name: str | None = None
    payment_expired_at: str | None = None
    payment_notification: bool = False
    phone: str | None = None
    profile_id: str | None = None
    proxy: str | None = None
    uuid: str | None = None
    webhook_types: tuple[str, ...] = ()
    webhook_url: str | None = None
    worked_days: int = 0
